#
# Copyright (C) 2011 Pedro Algarvio <pedro@algarvio.me>
#
# This file is part of Deluge and is licensed under GNU General Public License 3.0, or later, with
# the additional special exception to link portions of this program with the OpenSSL library.
# See LICENSE for more details.
#

from math import pi

import gi  # isort:skip (Version check required before import).

gi.require_version('PangoCairo', '1.0')
gi.require_foreign('cairo')
gi.require_version('cairo', '1.0')

# isort:imports-thirdparty
import cairo  # Backward compat cairo <= 1.15
from gi.repository import PangoCairo
from gi.repository.Gtk import DrawingArea, ProgressBar, StateFlags
from gi.repository.Pango import SCALE, Weight

# isort:imports-firstparty
from deluge.configmanager import ConfigManager

COLOR_STATES = ['missing', 'waiting', 'downloading', 'completed']


class PiecesBar(DrawingArea):
    # Draw in response to an draw
    __gsignals__ = {'draw': 'override'}

    def __init__(self):
        super().__init__()
        # Get progress bar styles, in order to keep font consistency
        pb = ProgressBar()
        pb_style = pb.get_style_context()
        # Get a copy of Pango.FontDescription since original needs freed.
        self.text_font = pb_style.get_property('font', StateFlags.NORMAL).copy()
        self.text_font.set_weight(Weight.BOLD)
        # Done with the ProgressBar styles, don't keep refs of it
        del pb, pb_style

        self.set_size_request(-1, 25)
        self.gtkui_config = ConfigManager('gtk3ui.conf')

        self.width = self.prev_width = 0
        self.height = self.prev_height = 0
        self.pieces = self.prev_pieces = ()
        self.num_pieces = None
        self.text = self.prev_text = ''
        self.fraction = self.prev_fraction = 0
        self.progress_overlay = self.text_overlay = self.pieces_overlay = None

        self.connect('size-allocate', self.do_size_allocate_event)
        self.show()

    def do_size_allocate_event(self, widget, size):
        self.prev_width = self.width
        self.width = size.width
        self.prev_height = self.height
        self.height = size.height

    # Handle the draw by drawing
    def do_draw(self, ctx):
        ctx.set_line_width(max(ctx.device_to_user_distance(0.5, 0.5)))

        # Restrict Cairo to the exposed area; avoid extra work
        self.roundcorners_clipping(ctx)

        self.draw_pieces(ctx)
        self.draw_progress_overlay(ctx)
        self.write_text(ctx)
        self.roundcorners_border(ctx)

        # Drawn once, update width, height
        if self.resized():
            self.prev_width = self.width
            self.prev_height = self.height

    def roundcorners_clipping(self, ctx):
        self.create_roundcorners_subpath(ctx, 0, 0, self.width, self.height)
        ctx.clip()

    def roundcorners_border(self, ctx):
        self.create_roundcorners_subpath(ctx, 0.5, 0.5, self.width - 1, self.height - 1)
        ctx.set_source_rgba(0, 0, 0, 0.9)
        ctx.stroke()

    @staticmethod
    def create_roundcorners_subpath(ctx, x, y, width, height):
        aspect = 1.0
        corner_radius = height / 10
        radius = corner_radius / aspect
        degrees = pi / 180
        ctx.new_sub_path()
        ctx.arc(x + width - radius, y + radius, radius, -90 * degrees, 0 * degrees)
        ctx.arc(
            x + width - radius, y + height - radius, radius, 0 * degrees, 90 * degrees
        )
        ctx.arc(x + radius, y + height - radius, radius, 90 * degrees, 180 * degrees)
        ctx.arc(x + radius, y + radius, radius, 180 * degrees, 270 * degrees)
        ctx.close_path()

    def draw_pieces(self, ctx):
        if not self.num_pieces:
            return

        if (
            self.resized()
            or self.pieces != self.prev_pieces
            or self.pieces_overlay is None
        ):
            # Need to recreate the cache drawing
            self.pieces_overlay = cairo.ImageSurface(
                cairo.FORMAT_ARGB32, self.width, self.height
            )
            pieces_ctx = cairo.Context(self.pieces_overlay)

            if self.pieces:
                pieces = self.pieces
            elif self.num_pieces:
                # Completed torrents do not send any pieces so create list using 'completed' state.
                pieces = [COLOR_STATES.index('completed')] * self.num_pieces
            start_pos = 0
            piece_width = self.width / len(pieces)
            pieces_colors = [
                [
                    color / 65535
                    for color in self.gtkui_config['pieces_color_%s' % state]
                ]
                for state in COLOR_STATES
            ]
            for state in pieces:
                pieces_ctx.set_source_rgb(*pieces_colors[state])
                pieces_ctx.rectangle(start_pos, 0, piece_width, self.height)
                pieces_ctx.fill()
                start_pos += piece_width

        ctx.set_source_surface(self.pieces_overlay)
        ctx.paint()

    def draw_progress_overlay(self, ctx):
        if not self.text:
            return

        if (
            self.resized()
            or self.fraction != self.prev_fraction
            or self.progress_overlay is None
        ):
            # Need to recreate the cache drawing
            self.progress_overlay = cairo.ImageSurface(
                cairo.FORMAT_ARGB32, self.width, self.height
            )
            progress_ctx = cairo.Context(self.progress_overlay)
            progress_ctx.set_source_rgba(0.1, 0.1, 0.1, 0.3)  # Transparent
            progress_ctx.rectangle(0, 0, self.width * self.fraction, self.height)
            progress_ctx.fill()
        ctx.set_source_surface(self.progress_overlay)
        ctx.paint()

    def write_text(self, ctx):
        if not self.text:
            return

        if self.resized() or self.text != self.prev_text or self.text_overlay is None:
            # Need to recreate the cache drawing
            self.text_overlay = cairo.ImageSurface(
                cairo.FORMAT_ARGB32, self.width, self.height
            )
            text_ctx = cairo.Context(self.text_overlay)
            pl = PangoCairo.create_layout(text_ctx)
            pl.set_font_description(self.text_font)
            pl.set_width(-1)  # No text wrapping
            pl.set_text(self.text, -1)
            plsize = pl.get_size()
            text_width = plsize[0] // SCALE
            text_height = plsize[1] // SCALE
            area_width_without_text = self.width - text_width
            area_height_without_text = self.height - text_height
            text_ctx.move_to(
                area_width_without_text // 2, area_height_without_text // 2
            )
            text_ctx.set_source_rgb(1, 1, 1)
            PangoCairo.update_layout(text_ctx, pl)
            PangoCairo.show_layout(text_ctx, pl)
        ctx.set_source_surface(self.text_overlay)
        ctx.paint()

    def resized(self):
        return self.prev_width != self.width or self.prev_height != self.height

    def set_fraction(self, fraction):
        self.prev_fraction = self.fraction
        self.fraction = fraction

    def get_fraction(self):
        return self.fraction

    def get_text(self):
        return self.text

    def set_text(self, text):
        self.prev_text = self.text
        self.text = text

    def set_pieces(self, pieces, num_pieces):
        self.prev_pieces = self.pieces
        self.pieces = pieces
        self.num_pieces = num_pieces

    def get_pieces(self):
        return self.pieces

    def clear(self):
        self.pieces = self.prev_pieces = ()
        self.num_pieces = None
        self.text = self.prev_text = ''
        self.fraction = self.prev_fraction = 0
        self.progress_overlay = self.text_overlay = self.pieces_overlay = None
        self.update()

    def update(self):
        self.queue_draw()



#
# Copyright (C) 2007 Andrew Resch <andrewresch@gmail.com>
#
# This file is part of Deluge and is licensed under GNU General Public License 3.0, or later, with
# the additional special exception to link portions of this program with the OpenSSL library.
# See LICENSE for more details.
#

import logging
import os.path

from gi.repository.GLib import timeout_add
from gi.repository.Gtk import Builder, CellRendererText, ListStore, TreeViewColumn

import deluge.common
import deluge.component as component
from deluge.configmanager import ConfigManager

from .common import get_logo
from .ipcinterface import process_args

log = logging.getLogger(__name__)


class QueuedTorrents(component.Component):
    def __init__(self):
        component.Component.__init__(
            self, 'QueuedTorrents', depend=['StatusBar', 'AddTorrentDialog']
        )
        self.queue = []
        self.status_item = None

        self.config = ConfigManager('gtk3ui.conf')
        self.builder = Builder()
        self.builder.add_from_file(
            deluge.common.resource_filename(
                __package__, os.path.join('glade', 'queuedtorrents.ui')
            )
        )
        self.builder.get_object('chk_autoadd').set_active(self.config['autoadd_queued'])
        self.dialog = self.builder.get_object('queued_torrents_dialog')
        self.dialog.set_icon(get_logo(32))

        self.builder.connect_signals(self)

        self.treeview = self.builder.get_object('treeview')
        self.treeview.append_column(
            TreeViewColumn(_('Torrent'), CellRendererText(), text=0)
        )

        self.liststore = ListStore(str, str)
        self.treeview.set_model(self.liststore)
        self.treeview.set_tooltip_column(1)

    def run(self):
        self.dialog.set_transient_for(component.get('MainWindow').window)
        self.dialog.show()

    def start(self):
        if len(self.queue) == 0:
            return

        # Make sure status bar info is showing
        self.update_status_bar()

        # We only want the add button sensitive if we're connected to a host
        self.builder.get_object('button_add').set_sensitive(True)

        if self.config['autoadd_queued'] or self.config['standalone']:
            self.on_button_add_clicked(None)
        else:
            self.run()

    def stop(self):
        # We only want the add button sensitive if we're connected to a host
        self.builder.get_object('button_add').set_sensitive(False)
        self.update_status_bar()

    def add_to_queue(self, torrents):
        """Adds the list of torrents to the queue"""
        # Add to the queue while removing duplicates
        self.queue = list(set(self.queue + torrents))

        # Update the liststore
        self.liststore.clear()
        for torrent in self.queue:
            if deluge.common.is_magnet(torrent):
                magnet = deluge.common.get_magnet_info(torrent)
                self.liststore.append([magnet['name'], torrent])
            else:
                self.liststore.append([os.path.split(torrent)[1], torrent])

        # Update the status bar
        self.update_status_bar()

    def update_status_bar(self):
        """Attempts to update status bar"""
        # If there are no queued torrents.. remove statusbar widgets and return
        if len(self.queue) == 0:
            if self.status_item is not None:
                component.get('StatusBar').remove_item(self.status_item)
                self.status_item = None
            return False

        try:
            component.get('StatusBar')
        except Exception:
            # The statusbar hasn't been loaded yet, so we'll add a timer to
            # update it later.
            timeout_add(100, self.update_status_bar)
            return False

        # Set the label text for statusbar
        if len(self.queue) > 1:
            label = str(len(self.queue)) + _(' Torrents Queued')
        else:
            label = str(len(self.queue)) + _(' Torrent Queued')

        # Add the statusbar items if needed, or just modify the label if they
        # have already been added.
        if self.status_item is None:
            self.status_item = component.get('StatusBar').add_item(
                icon='view-sort-descending',
                text=label,
                callback=self.on_statusbar_click,
            )
        else:
            self.status_item.set_text(label)

        # We return False so the timer stops
        return False

    def on_statusbar_click(self, widget, event):
        log.debug('on_statusbar_click')
        self.run()

    def on_button_remove_clicked(self, widget):
        selected = self.treeview.get_selection().get_selected()[1]
        if selected is not None:
            path = self.liststore.get_value(selected, 1)
            self.liststore.remove(selected)
            self.queue.remove(path)
            self.update_status_bar()

    def on_button_clear_clicked(self, widget):
        self.liststore.clear()
        del self.queue[:]
        self.update_status_bar()

    def on_button_close_clicked(self, widget):
        self.dialog.hide()

    def on_button_add_clicked(self, widget):
        # Add all the torrents in the liststore
        def add_torrent(model, path, _iter, data):
            torrent_path = deluge.common.decode_bytes(model.get_value(_iter, 1))
            process_args([torrent_path])

        self.liststore.foreach(add_torrent, None)
        del self.queue[:]
        self.dialog.hide()
        self.update_status_bar()

    def on_chk_autoadd_toggled(self, widget):
        self.config['autoadd_queued'] = widget.get_active()



#
# Copyright (C) 2007 Andrew Resch <andrewresch@gmail.com>
# Copyright (C) 2008 Martijn Voncken <mvoncken@gmail.com>
#
# This file is part of Deluge and is licensed under GNU General Public License 3.0, or later, with
# the additional special exception to link portions of this program with the OpenSSL library.
# See LICENSE for more details.
#

import logging

from gi.repository.Gtk import Label, PolicyType, ScrolledWindow

import deluge.component as component
from deluge.configmanager import ConfigManager

log = logging.getLogger(__name__)


class SideBar(component.Component):
    """
    manages the sidebar-tabs.
    purpose : plugins
    """

    def __init__(self):
        component.Component.__init__(self, 'SideBar')
        main_builder = component.get('MainWindow').get_builder()
        self.notebook = main_builder.get_object('sidebar_notebook')
        self.config = ConfigManager('gtk3ui.conf')

        # Tabs holds references to the Tab widgets by their name
        self.tabs = {}

        # Hide if necessary
        self.visible(self.config['show_sidebar'])

    def visible(self, visible):
        self.notebook.show() if visible else self.notebook.hide()
        self.config['show_sidebar'] = visible

    def add_tab(self, widget, tab_name, label):
        """Adds a tab object to the notebook."""
        log.debug('add tab: %s', tab_name)
        self.tabs[tab_name] = widget
        scrolled = ScrolledWindow()
        scrolled.set_policy(PolicyType.AUTOMATIC, PolicyType.AUTOMATIC)
        scrolled.add(widget)
        self.notebook.insert_page(scrolled, Label(label=label), -1)
        scrolled.show_all()

        self.after_update()

    def remove_tab(self, tab_name):
        """Removes a tab by name."""
        self.notebook.remove_page(self.notebook.page_num(self.tabs[tab_name]))
        del self.tabs[tab_name]

        self.after_update()

    def after_update(self):
        # If there are no tabs visible, then do not show the notebook
        if len(self.tabs) == 0:
            self.visible(False)

        # If there is 1 tab, hide the tab-headers
        if len(self.tabs) == 1:
            self.notebook.set_show_tabs(False)
        else:
            self.notebook.set_show_tabs(True)



#
# Copyright (C) 2009 Andrew Resch <andrewresch@gmail.com>
#
# This file is part of Deluge and is licensed under GNU General Public License 3.0, or later, with
# the additional special exception to link portions of this program with the OpenSSL library.
# See LICENSE for more details.
#

# pylint: disable=super-on-old-class

from collections import namedtuple

from gi.repository import Gtk
from twisted.internet import defer

import deluge.component as component
from deluge.common import windows_check

from .common import get_deluge_icon, get_pixbuf

Account = namedtuple('Account', 'username password authlevel')


class BaseDialog(Gtk.Dialog):
    """
    Base dialog class that should be used with all dialogs.
    """

    def __init__(self, header, text, icon, buttons, parent=None):
        """
        :param header: str, the header portion of the dialog
        :param text: str, the text body of the dialog
        :param icon: icon name from icon theme or icon filename.
        :param buttons: tuple, of icon name and responses
        :param parent: gtkWindow, the parent window, if None it will default to the
            MainWindow
        """
        super().__init__(
            title=header,
            parent=parent if parent else component.get('MainWindow').window,
            flags=Gtk.DialogFlags.MODAL | Gtk.DialogFlags.DESTROY_WITH_PARENT,
            buttons=buttons,
        )

        self.set_icon(get_deluge_icon())

        self.connect('delete-event', self._on_delete_event)
        self.connect('response', self._on_response)

        # Setup all the formatting and such to make our dialog look pretty
        self.set_border_width(5)
        self.set_default_size(200, 100)
        hbox = Gtk.Box.new(Gtk.Orientation.HORIZONTAL, spacing=5)
        image = Gtk.Image()
        if icon.endswith('.svg') or icon.endswith('.png'):
            # Hack for Windows since it doesn't support svg
            if icon.endswith('.svg') and windows_check():
                icon = icon.rpartition('.svg')[0] + '16.png'
            image.set_from_pixbuf(get_pixbuf(icon, 24))
        else:
            image.set_from_icon_name(icon, Gtk.IconSize.LARGE_TOOLBAR)
        image.set_alignment(0.5, 0.0)
        hbox.pack_start(image, False, False, 0)
        vbox = Gtk.Box.new(Gtk.Orientation.VERTICAL, spacing=5)
        tlabel = Gtk.Label(label=text)
        tlabel.set_use_markup(True)
        tlabel.set_line_wrap(True)
        tlabel.set_alignment(0.0, 0.5)
        vbox.pack_start(tlabel, False, False, 0)
        hbox.pack_start(vbox, False, False, 0)
        self.vbox.pack_start(hbox, False, False, 0)
        self.vbox.set_spacing(5)
        self.vbox.show_all()

    def _on_delete_event(self, widget, event):
        self.destroy()
        self.deferred.callback(Gtk.ResponseType.DELETE_EVENT)

    def _on_response(self, widget, response):
        self.destroy()
        self.deferred.callback(response)

    def run(self):
        """
        Shows the dialog and returns a Deferred object.  The deferred, when fired
        will contain the response ID.
        """
        self.deferred = defer.Deferred()
        self.show()
        return self.deferred


class YesNoDialog(BaseDialog):
    """
    Displays a dialog asking the user to select Yes or No to a question.

    When run(), it will return either a Gtk.ResponseType.YES or a Gtk.ResponseType.NO.

    """

    def __init__(self, header, text, parent=None):
        """
        :param header: see `:class:BaseDialog`
        :param text: see `:class:BaseDialog`
        :param parent: see `:class:BaseDialog`
        """
        super().__init__(
            header,
            text,
            'dialog-question',
            (_('_No'), Gtk.ResponseType.NO, _('_Yes'), Gtk.ResponseType.YES),
            parent,
        )
        # Use the preferred size calculated from the content
        self.set_default_size(-1, -1)


class InformationDialog(BaseDialog):
    """
    Displays an information dialog.

    When run(), it will return a Gtk.ResponseType.CLOSE.
    """

    def __init__(self, header, text, parent=None):
        """
        :param header: see `:class:BaseDialog`
        :param text: see `:class:BaseDialog`
        :param parent: see `:class:BaseDialog`
        """
        super().__init__(
            header,
            text,
            'dialog-information',
            (_('_Close'), Gtk.ResponseType.CLOSE),
            parent,
        )


class ErrorDialog(BaseDialog):
    """
    Displays an error dialog with optional details text for more information.

    When run(), it will return a Gtk.ResponseType.CLOSE.
    """

    def __init__(self, header, text, parent=None, details=None, traceback=False):
        """
        :param header: see `:class:BaseDialog`
        :param text: see `:class:BaseDialog`
        :param parent: see `:class:BaseDialog`
        :param details: extra information that will be displayed in a
            scrollable textview
        :type details: string
        :param traceback: show the traceback information in the details area
        :type traceback: bool
        """
        super().__init__(
            header, text, 'dialog-error', (_('_Close'), Gtk.ResponseType.CLOSE), parent
        )

        if traceback:
            import sys
            import traceback

            tb = sys.exc_info()
            tb = traceback.format_exc(tb[2])
            if details:
                details += '\n' + tb
            else:
                details = tb

        if details:
            self.set_default_size(600, 400)
            textview = Gtk.TextView()
            textview.set_editable(False)
            textview.get_buffer().set_text(details)
            sw = Gtk.ScrolledWindow()
            sw.set_policy(Gtk.PolicyType.AUTOMATIC, Gtk.PolicyType.AUTOMATIC)
            sw.set_shadow_type(Gtk.ShadowType.IN)
            sw.add(textview)
            label = Gtk.Label(label=_('Details:'))
            label.set_alignment(0.0, 0.5)
            self.vbox.pack_start(label, False, False, 0)
            self.vbox.pack_start(sw, True, True, 0)
            self.vbox.show_all()


class AuthenticationDialog(BaseDialog):
    """
    Displays a dialog with entry fields asking for username and password.

    When run(), it will return either a Gtk.ResponseType.CANCEL or a
    Gtk.ResponseType.OK.
    """

    def __init__(self, err_msg='', username=None, parent=None):
        """
        :param err_msg: the error message we got back from the server
        :type err_msg: string
        """
        super().__init__(
            _('Authenticate'),
            err_msg,
            'dialog-password',
            (_('_Cancel'), Gtk.ResponseType.CANCEL, _('C_onnect'), Gtk.ResponseType.OK),
            parent,
        )

        table = Gtk.Table(2, 2, False)
        self.username_label = Gtk.Label()
        self.username_label.set_markup('<b>' + _('Username:') + '</b>')
        self.username_label.set_alignment(1.0, 0.5)
        self.username_label.set_padding(5, 5)
        self.username_entry = Gtk.Entry()
        table.attach(self.username_label, 0, 1, 0, 1)
        table.attach(self.username_entry, 1, 2, 0, 1)

        self.password_label = Gtk.Label()
        self.password_label.set_markup('<b>' + _('Password:') + '</b>')
        self.password_label.set_alignment(1.0, 0.5)
        self.password_label.set_padding(5, 5)
        self.password_entry = Gtk.Entry()
        self.password_entry.set_visibility(False)
        self.password_entry.connect('activate', self.on_password_activate)
        table.attach(self.password_label, 0, 1, 1, 2)
        table.attach(self.password_entry, 1, 2, 1, 2)

        self.vbox.pack_start(table, False, False, padding=5)
        self.set_focus(self.password_entry)
        if username:
            self.username_entry.set_text(username)
            self.username_entry.set_editable(False)
            self.set_focus(self.password_entry)
        else:
            self.set_focus(self.username_entry)
        self.show_all()

    def get_username(self):
        return self.username_entry.get_text()

    def get_password(self):
        return self.password_entry.get_text()

    def on_password_activate(self, widget):
        self.response(Gtk.ResponseType.OK)


class AccountDialog(BaseDialog):
    def __init__(
        self,
        username=None,
        password=None,
        authlevel=None,
        levels_mapping=None,
        parent=None,
    ):
        if username:
            super().__init__(
                _('Edit Account'),
                _('Edit existing account'),
                'dialog-information',
                (
                    _('_Cancel'),
                    Gtk.ResponseType.CANCEL,
                    _('_Apply'),
                    Gtk.ResponseType.OK,
                ),
                parent,
            )
        else:
            super().__init__(
                _('New Account'),
                _('Create a new account'),
                'dialog-information',
                (_('_Cancel'), Gtk.ResponseType.CANCEL, _('_Add'), Gtk.ResponseType.OK),
                parent,
            )

        self.account = None

        table = Gtk.Table(2, 3, False)
        username_label = Gtk.Label()
        username_label.set_markup('<b>' + _('Username:') + '</b>')
        username_label.set_alignment(1.0, 0.5)
        username_label.set_padding(5, 5)
        self.username_entry = Gtk.Entry()
        table.attach(username_label, 0, 1, 0, 1)
        table.attach(self.username_entry, 1, 2, 0, 1)

        authlevel_label = Gtk.Label()
        authlevel_label.set_markup('<b>' + _('Authentication Level:') + '</b>')
        authlevel_label.set_alignment(1.0, 0.5)
        authlevel_label.set_padding(5, 5)

        # combo_box_new_text is deprecated but no other pygtk alternative.
        self.authlevel_combo = Gtk.ComboBoxText()
        active_idx = None
        for idx, level in enumerate(levels_mapping):
            self.authlevel_combo.append_text(level)
            if authlevel and authlevel == level:
                active_idx = idx
            elif not authlevel and level == 'DEFAULT':
                active_idx = idx

        if active_idx is not None:
            self.authlevel_combo.set_active(active_idx)

        table.attach(authlevel_label, 0, 1, 1, 2)
        table.attach(self.authlevel_combo, 1, 2, 1, 2)

        password_label = Gtk.Label()
        password_label.set_markup('<b>' + _('Password:') + '</b>')
        password_label.set_alignment(1.0, 0.5)
        password_label.set_padding(5, 5)
        self.password_entry = Gtk.Entry()
        self.password_entry.set_visibility(False)
        table.attach(password_label, 0, 1, 2, 3)
        table.attach(self.password_entry, 1, 2, 2, 3)

        self.vbox.pack_start(table, False, False, padding=5)
        if username:
            self.username_entry.set_text(username)
            self.username_entry.set_editable(False)
        else:
            self.set_focus(self.username_entry)

        if password:
            self.password_entry.set_text(username)

        self.vbox.show_all()

    def _on_response(self, widget, response):
        if response == Gtk.ResponseType.OK:
            self.account = Account(
                self.username_entry.get_text(),
                self.password_entry.get_text(),
                self.authlevel_combo.get_active_text(),
            )
        self.destroy()
        self.deferred.callback(response)


class OtherDialog(BaseDialog):
    """
    Displays a dialog with a spinner for setting a value.

    Returns:
        int or float:
    """

    def __init__(
        self, header, text='', unit_text='', icon=None, default=0, parent=None
    ):
        self.value_type = type(default)
        if self.value_type not in (int, float):
            raise TypeError('default value needs to be an int or float')

        if not icon:
            icon = 'dialog-information'

        super().__init__(
            header,
            text,
            icon,
            (_('_Cancel'), Gtk.ResponseType.CANCEL, _('_Apply'), Gtk.ResponseType.OK),
            parent,
        )

        hbox = Gtk.Box.new(Gtk.Orientation.HORIZONTAL, spacing=5)
        alignment_spacer = Gtk.Alignment()
        hbox.pack_start(alignment_spacer, True, True, 0)
        alignment_spin = Gtk.Alignment(xalign=1, yalign=0.5, xscale=1, yscale=1)
        adjustment_spin = Gtk.Adjustment(
            value=-1, lower=-1, upper=2097151, step_increment=1, page_increment=10
        )
        self.spinbutton = Gtk.SpinButton(
            adjustment=adjustment_spin, climb_rate=0, digits=0
        )
        self.spinbutton.set_value(default)
        self.spinbutton.select_region(0, -1)
        self.spinbutton.set_width_chars(6)
        self.spinbutton.set_alignment(1)
        self.spinbutton.set_max_length(6)
        if self.value_type is float:
            self.spinbutton.set_digits(1)
        alignment_spin.add(self.spinbutton)
        hbox.pack_start(alignment_spin, False, True, 0)
        label_type = Gtk.Label()
        label_type.set_text(unit_text)
        label_type.set_alignment(0.0, 0.5)
        hbox.pack_start(label_type, True, True, 0)

        self.vbox.pack_start(hbox, False, False, padding=5)
        self.vbox.show_all()

    def _on_delete_event(self, widget, event):
        self.deferred.callback(None)
        self.destroy()

    def _on_response(self, widget, response):
        value = None
        if response == Gtk.ResponseType.OK:
            if self.value_type is int:
                value = self.spinbutton.get_value_as_int()
            else:
                value = self.spinbutton.get_value()
        self.deferred.callback(value)
        self.destroy()


class PasswordDialog(BaseDialog):
    """
    Displays a dialog with an entry field asking for a password.

    When run(), it will return either a Gtk.ResponseType.CANCEL or a Gtk.ResponseType.OK.
    """

    def __init__(self, password_msg='', parent=None):
        """
        :param password_msg: the error message we got back from the server
        :type password_msg: string
        """
        super().__init__(
            header=_('Password Protected'),
            text=password_msg,
            icon='dialog-password',
            buttons=(
                _('_Cancel'),
                Gtk.ResponseType.CANCEL,
                _('_OK'),
                Gtk.ResponseType.OK,
            ),
            parent=parent,
        )

        table = Gtk.Table(1, 2, False)
        self.password_label = Gtk.Label()
        self.password_label.set_markup('<b>' + _('Password:') + '</b>')
        self.password_label.set_alignment(1.0, 0.5)
        self.password_label.set_padding(5, 5)
        self.password_entry = Gtk.Entry()
        self.password_entry.set_visibility(False)
        self.password_entry.connect('activate', self.on_password_activate)
        table.attach(self.password_label, 0, 1, 1, 2)
        table.attach(self.password_entry, 1, 2, 1, 2)

        self.vbox.pack_start(table, False, False, padding=5)
        self.set_focus(self.password_entry)

        self.show_all()

    def get_password(self):
        return self.password_entry.get_text()

    def on_password_activate(self, widget):
        self.response(Gtk.ResponseType.OK)


class CopyMagnetDialog(BaseDialog):
    """
    Displays a dialog with a magnet URI
    """

    def __init__(self, torrent_magnet='', parent=None):
        super().__init__(
            header=_('Copy Magnet URI'),
            text='',
            icon='magnet_copy.svg',
            buttons=(_('_Close'), Gtk.ResponseType.CLOSE),
            parent=parent,
        )
        self.copied = False

        table = Gtk.Table(1, 2, False)
        self.magnet_entry = Gtk.Entry()
        self.magnet_entry.set_text(torrent_magnet)
        self.magnet_entry.set_editable(False)
        self.magnet_entry.connect('copy-clipboard', self.on_copy_emitted)
        table.attach(self.magnet_entry, 0, 1, 0, 1)

        copy_button = Gtk.Button.new_with_label(_('Copy'))
        copy_button.connect('clicked', self.on_copy_clicked)
        table.attach(copy_button, 1, 2, 0, 1)

        self.vbox.pack_start(table, False, False, padding=5)
        self.set_focus(self.magnet_entry)

        self.show_all()

    def on_copy_clicked(self, widget):
        self.magnet_entry.select_region(0, -1)
        self.magnet_entry.copy_clipboard()
        self.magnet_entry.set_position(0)
        self.copied = True

    def on_copy_emitted(self, widget):
        self.copied = True



#
# Copyright (C) 2008 Andrew Resch <andrewresch@gmail.com>
#
# This file is part of Deluge and is licensed under GNU General Public License 3.0, or later, with
# the additional special exception to link portions of this program with the OpenSSL library.
# See LICENSE for more details.
#

import json
import logging
import os.path

import gi  # isort:skip (Required before Gtk import).

gi.require_version('Gtk', '3.0')

# isort:imports-thirdparty
from gi.repository import Gio, Gtk
from gi.repository.Gdk import DragAction, ModifierType, keyval_name
from gi.repository.GObject import TYPE_UINT64

# isort:imports-firstparty
import deluge.component as component
from deluge.common import open_file, show_file
from deluge.ui.client import client
from deluge.ui.common import FILE_PRIORITY

# isort:imports-localfolder
from .common import (
    listview_replace_treestore,
    load_pickled_state_file,
    reparent_iter,
    save_pickled_state_file,
)
from .torrentdetails import Tab
from .torrentview_data_funcs import cell_data_size

log = logging.getLogger(__name__)

CELL_PRIORITY_ICONS = {
    FILE_PRIORITY['Skip']: 'action-unavailable-symbolic',
    FILE_PRIORITY['Low']: 'go-down-symbolic',
    FILE_PRIORITY['Normal']: 'go-next-symbolic',
    FILE_PRIORITY['High']: 'go-up-symbolic',
}

G_ICON_DIRECTORY = Gio.content_type_get_icon('inode/directory')


def cell_priority(column, cell, model, row, data):
    if model.get_value(row, 5) == -1:
        # This is a folder, so lets just set it blank for now
        cell.set_property('text', '')
        return
    priority = model.get_value(row, data)
    cell.set_property('text', _(FILE_PRIORITY[priority]))


def cell_priority_icon(column, cell, model, row, data):
    if model.get_value(row, 5) == -1:
        # This is a folder, so lets just set it blank for now
        cell.set_property('icon-name', None)
        return
    priority = model.get_value(row, data)
    cell.set_property('icon-name', CELL_PRIORITY_ICONS[priority])


def cell_filename(column, cell, model, row, data):
    """Only show the tail portion of the file path"""
    filepath = model.get_value(row, data)
    cell.set_property('text', os.path.split(filepath)[1])


def cell_progress(column, cell, model, row, data):
    text = model.get_value(row, data[0])
    value = model.get_value(row, data[1])
    cell.set_property('visible', True)
    cell.set_property('text', text)
    cell.set_property('value', value)


class FilesTab(Tab):
    def __init__(self):
        super().__init__('Files', 'files_tab', 'files_tab_label')

        self.listview = self.main_builder.get_object('files_listview')
        # filename, size, progress string, progress value, priority, file index, icon id
        self.treestore = Gtk.TreeStore(str, TYPE_UINT64, str, float, int, int, Gio.Icon)
        self.treestore.set_sort_column_id(0, Gtk.SortType.ASCENDING)

        # We need to store the row that's being edited to prevent updating it until
        # it's been done editing
        self._editing_index = None

        # Filename column
        self.filename_column_name = _('Filename')
        column = Gtk.TreeViewColumn(self.filename_column_name)
        render = Gtk.CellRendererPixbuf()
        column.pack_start(render, False)
        column.add_attribute(render, 'gicon', 6)
        render = Gtk.CellRendererText()
        render.set_property('editable', True)
        render.connect('edited', self._on_filename_edited)
        render.connect('editing-started', self._on_filename_editing_start)
        render.connect('editing-canceled', self._on_filename_editing_canceled)
        column.pack_start(render, True)
        column.add_attribute(render, 'text', 0)
        column.set_sort_column_id(0)
        column.set_clickable(True)
        column.set_resizable(True)
        column.set_expand(False)
        column.set_min_width(200)
        column.set_reorderable(True)
        self.listview.append_column(column)

        # Size column
        column = Gtk.TreeViewColumn(_('Size'))
        render = Gtk.CellRendererText()
        column.pack_start(render, False)
        column.set_cell_data_func(render, cell_data_size, 1)
        column.set_sort_column_id(1)
        column.set_clickable(True)
        column.set_resizable(True)
        column.set_expand(False)
        column.set_min_width(50)
        column.set_reorderable(True)
        self.listview.append_column(column)

        # Progress column
        column = Gtk.TreeViewColumn(_('Progress'))
        render = Gtk.CellRendererProgress()
        render.set_padding(0, 1)
        column.pack_start(render, True)
        column.set_cell_data_func(render, cell_progress, (2, 3))
        column.set_sort_column_id(3)
        column.set_clickable(True)
        column.set_resizable(True)
        column.set_expand(False)
        column.set_min_width(100)
        column.set_reorderable(True)
        self.listview.append_column(column)

        # Priority column
        column = Gtk.TreeViewColumn(_('Priority'))
        render = Gtk.CellRendererPixbuf()
        column.pack_start(render, False)
        column.set_cell_data_func(render, cell_priority_icon, 4)
        render = Gtk.CellRendererText()
        column.pack_start(render, False)
        column.set_cell_data_func(render, cell_priority, 4)
        column.set_sort_column_id(4)
        column.set_clickable(True)
        column.set_resizable(True)
        column.set_expand(False)
        column.set_min_width(100)
        # Bugfix: Last column needs max_width set to stop scrollbar appearing
        column.set_max_width(200)
        column.set_reorderable(True)
        self.listview.append_column(column)

        self.listview.set_model(self.treestore)

        self.listview.get_selection().set_mode(Gtk.SelectionMode.MULTIPLE)

        self.file_menu = self.main_builder.get_object('menu_file_tab')
        self.file_menu_priority_items = [
            self.main_builder.get_object('menuitem_skip'),
            self.main_builder.get_object('menuitem_low'),
            self.main_builder.get_object('menuitem_normal'),
            self.main_builder.get_object('menuitem_high'),
            self.main_builder.get_object('menuitem_priority_sep'),
        ]

        self.localhost_widgets = [
            self.main_builder.get_object('menuitem_open_file'),
            self.main_builder.get_object('menuitem_show_file'),
            self.main_builder.get_object('menuitem3'),
        ]

        self.listview.connect('row-activated', self._on_row_activated)
        self.listview.connect('key-press-event', self._on_key_press_event)
        self.listview.connect('button-press-event', self._on_button_press_event)

        self.listview.enable_model_drag_source(
            ModifierType.BUTTON1_MASK,
            [('text/plain', 0, 0)],
            DragAction.DEFAULT | DragAction.MOVE,
        )
        self.listview.enable_model_drag_dest([('text/plain', 0, 0)], DragAction.DEFAULT)

        self.listview.connect('drag_data_get', self._on_drag_data_get_data)
        self.listview.connect('drag_data_received', self._on_drag_data_received_data)

        component.get('MainWindow').connect_signals(self)

        # Connect to various events from the daemon
        client.register_event_handler(
            'TorrentFileRenamedEvent', self._on_torrentfilerenamed_event
        )
        client.register_event_handler(
            'TorrentFolderRenamedEvent', self._on_torrentfolderrenamed_event
        )
        client.register_event_handler(
            'TorrentRemovedEvent', self._on_torrentremoved_event
        )

        # Attempt to load state
        self.load_state()

        # torrent_id: (filepath, size)
        self.files_list = {}

        self.torrent_id = None

    def start(self):
        attr = 'hide' if not client.is_localhost() else 'show'
        for widget in self.localhost_widgets:
            getattr(widget, attr)()

    def save_state(self):
        # Get the current sort order of the view
        column_id, sort_order = self.treestore.get_sort_column_id()

        # Setup state dict
        state = {
            'columns': {},
            'sort_id': int(column_id) if column_id >= 0 else None,
            'sort_order': int(sort_order) if sort_order >= 0 else None,
        }

        for index, column in enumerate(self.listview.get_columns()):
            state['columns'][column.get_title()] = {
                'position': index,
                'width': column.get_width(),
            }

        save_pickled_state_file('files_tab.state', state)

    def load_state(self):
        state = load_pickled_state_file('files_tab.state')

        if not state:
            return

        if state['sort_id'] is not None and state['sort_order'] is not None:
            self.treestore.set_sort_column_id(state['sort_id'], state['sort_order'])

        for index, column in enumerate(self.listview.get_columns()):
            cname = column.get_title()
            if cname in state['columns']:
                cstate = state['columns'][cname]
                column.set_sizing(Gtk.TreeViewColumnSizing.FIXED)
                column.set_fixed_width(cstate['width'] if cstate['width'] > 0 else 10)
                if state['sort_id'] == index and state['sort_order'] is not None:
                    column.set_sort_indicator(True)
                    column.set_sort_order(state['sort_order'])
                if cstate['position'] != index:
                    # Column is in wrong position
                    if cstate['position'] == 0:
                        self.listview.move_column_after(column, None)
                    elif (
                        self.listview.get_columns()[cstate['position'] - 1].get_title()
                        != cname
                    ):
                        self.listview.move_column_after(
                            column, self.listview.get_columns()[cstate['position'] - 1]
                        )

    def update(self):
        # Get the first selected torrent
        torrent_id = component.get('TorrentView').get_selected_torrents()

        # Only use the first torrent in the list or return if None selected
        if len(torrent_id) != 0:
            torrent_id = torrent_id[0]
        else:
            # No torrent is selected in the torrentview
            self.clear()
            return

        status_keys = ['file_progress', 'file_priorities']
        if torrent_id != self.torrent_id:
            # We only want to do this if the torrent_id has changed
            self.treestore.clear()
            self.torrent_id = torrent_id
            status_keys += ['storage_mode', 'is_seed']

            if self.torrent_id in self.files_list:
                # We already have the files list stored, so just update the view
                self.update_files()

        if (
            self.torrent_id not in self.files_list
            or not self.files_list[self.torrent_id]
        ):
            # We need to get the files list
            log.debug('Getting file list from core..')
            status_keys += ['files']

        component.get('SessionProxy').get_torrent_status(
            self.torrent_id, status_keys
        ).addCallback(self._on_get_torrent_status, self.torrent_id)

    def clear(self):
        self.treestore.clear()
        self.torrent_id = None

    def _on_row_activated(self, tree, path, view_column):
        self.on_menuitem_open_file_activate(None)

    def get_file_path(self, row, path=''):
        if not row:
            return path

        path = self.treestore.get_value(row, 0) + path
        return self.get_file_path(self.treestore.iter_parent(row), path)

    def _on_open_file(self, status):
        paths = self.listview.get_selection().get_selected_rows()[1]
        selected = []
        for path in paths:
            selected.append(self.treestore.get_iter(path))

        for select in selected:
            path = self.get_file_path(select).split('/')
            filepath = os.path.join(status['download_location'], *path)
            log.debug('Open file: %s', filepath)
            timestamp = component.get('MainWindow').get_timestamp()
            open_file(filepath, timestamp=timestamp)

    def _on_show_file(self, status):
        paths = self.listview.get_selection().get_selected_rows()[1]
        selected = []
        for path in paths:
            selected.append(self.treestore.get_iter(path))

        for select in selected:
            path = self.get_file_path(select).split('/')
            filepath = os.path.join(status['download_location'], *path)
            log.debug('Show file: %s', filepath)
            timestamp = component.get('MainWindow').get_timestamp()
            show_file(filepath, timestamp=timestamp)

    # The following 3 methods create the folder/file view in the treeview
    def prepare_file_store(self, torrent_files):
        split_files = {}
        for index, torrent_file in enumerate(torrent_files):
            self.prepare_file(torrent_file, torrent_file['path'], index, split_files)
        self.add_files(None, split_files)

    def prepare_file(self, torrent_file, file_name, file_num, files_storage):
        first_slash_index = file_name.find('/')
        if first_slash_index == -1:
            files_storage[file_name] = (file_num, torrent_file)
        else:
            file_name_chunk = file_name[: first_slash_index + 1]
            if file_name_chunk not in files_storage:
                files_storage[file_name_chunk] = {}
            self.prepare_file(
                torrent_file,
                file_name[first_slash_index + 1 :],
                file_num,
                files_storage[file_name_chunk],
            )

    def add_files(self, parent_iter, split_files):
        chunk_size_total = 0
        for key, value in split_files.items():
            if key.endswith('/'):
                chunk_iter = self.treestore.append(
                    parent_iter, [key, 0, '', 0, 0, -1, G_ICON_DIRECTORY]
                )
                chunk_size = self.add_files(chunk_iter, value)
                self.treestore.set(chunk_iter, 1, chunk_size)
                chunk_size_total += chunk_size
            else:
                mime_type, uncertain = Gio.content_type_guess(key, None)
                if not uncertain and mime_type:
                    mime_icon = Gio.content_type_get_symbolic_icon(mime_type)
                else:
                    mime_icon = Gio.content_type_get_symbolic_icon('text/plain')
                self.treestore.append(
                    parent_iter, [key, value[1]['size'], '', 0, 0, value[0], mime_icon]
                )
                chunk_size_total += value[1]['size']
        return chunk_size_total

    def update_files(self):
        with listview_replace_treestore(self.listview):
            self.prepare_file_store(self.files_list[self.torrent_id])
        root = Gtk.TreePath.new_first()
        self.listview.expand_row(root, False)

    def get_selected_files(self):
        """Returns a list of file indexes that are selected."""

        def get_iter_children(itr, selected):
            i = self.treestore.iter_children(itr)
            while i:
                selected.append(self.treestore[i][5])
                if self.treestore.iter_has_child(i):
                    get_iter_children(i, selected)
                i = self.treestore.iter_next(i)

        selected = []
        paths = self.listview.get_selection().get_selected_rows()[1]
        for path in paths:
            i = self.treestore.get_iter(path)
            selected.append(self.treestore[i][5])
            if self.treestore.iter_has_child(i):
                get_iter_children(i, selected)

        return selected

    def get_files_from_tree(self, rows, files_list, indent):
        if not rows:
            return None

        for row in rows:
            if row[5] > -1:
                files_list.append((row[5], row))
            self.get_files_from_tree(row.iterchildren(), files_list, indent + 1)
        return None

    def update_folder_percentages(self):
        """Go through the tree and update the folder complete percentages."""
        root = self.treestore.get_iter_first()
        if root is None or self.treestore[root][5] != -1:
            return

        def get_completed_bytes(row):
            completed_bytes = 0
            parent = self.treestore.iter_parent(row)
            while row:
                if self.treestore.iter_children(row):
                    completed_bytes += get_completed_bytes(
                        self.treestore.iter_children(row)
                    )
                else:
                    completed_bytes += (
                        self.treestore[row][1] * self.treestore[row][3] / 100
                    )

                row = self.treestore.iter_next(row)

            try:
                value = completed_bytes / self.treestore[parent][1] * 100
            except ZeroDivisionError:
                # Catch the unusual error found when moving folders around
                value = 0
            self.treestore[parent][3] = value
            self.treestore[parent][2] = '%i%%' % value
            return completed_bytes

        get_completed_bytes(self.treestore.iter_children(root))

    def _on_get_torrent_status(self, status, torrent_id):
        # Check stored torrent id matches the callback id
        if self.torrent_id != torrent_id:
            return

        if 'is_seed' in status:
            self.__is_seed = status['is_seed']

        if 'files' in status:
            self.files_list[self.torrent_id] = status['files']
            self.update_files()

        # (index, iter)
        files_list = []
        self.get_files_from_tree(self.treestore, files_list, 0)
        files_list.sort()
        for index, row in files_list:
            # Do not update a row that is being edited
            if self._editing_index == row[5]:
                continue

            try:
                progress_string = '%i%%' % (status['file_progress'][index] * 100)
            except IndexError:
                continue
            if row[2] != progress_string:
                row[2] = progress_string
            progress_value = status['file_progress'][index] * 100
            if row[3] != progress_value:
                row[3] = progress_value
            file_priority = status['file_priorities'][index]
            if row[4] != file_priority:
                row[4] = file_priority
        if self._editing_index != -1:
            # Only update if no folder is being edited
            self.update_folder_percentages()

    def _on_button_press_event(self, widget, event):
        """This is a callback for showing the right-click context menu."""
        log.debug('on_button_press_event')
        # We only care about right-clicks
        if event.button == 3:
            x, y = event.get_coords()
            cursor_path = self.listview.get_path_at_pos(int(x), int(y))
            if not cursor_path:
                return

            paths = self.listview.get_selection().get_selected_rows()[1]
            if cursor_path[0] not in paths:
                row = self.treestore.get_iter(cursor_path[0])
                self.listview.get_selection().unselect_all()
                self.listview.get_selection().select_iter(row)

            for widget in self.file_menu_priority_items:
                widget.set_sensitive(not self.__is_seed)

            self.file_menu.popup(None, None, None, None, event.button, event.time)
            return True

    def _on_key_press_event(self, widget, event):
        keyname = keyval_name(event.keyval)
        if keyname is not None:
            func = getattr(self, 'keypress_' + keyname.lower(), None)
            selected_rows = self.listview.get_selection().get_selected_rows()[1]
            if func and selected_rows:
                return func(event)

    def keypress_menu(self, event):
        self.file_menu.popup(None, None, None, None, 3, event.time)
        return True

    def keypress_f2(self, event):
        path, col = self.listview.get_cursor()
        for column in self.listview.get_columns():
            if column.get_title() == self.filename_column_name:
                self.listview.set_cursor(path, column, True)
                return True

    def on_menuitem_open_file_activate(self, menuitem):
        if client.is_localhost:
            component.get('SessionProxy').get_torrent_status(
                self.torrent_id, ['download_location']
            ).addCallback(self._on_open_file)

    def on_menuitem_show_file_activate(self, menuitem):
        if client.is_localhost:
            component.get('SessionProxy').get_torrent_status(
                self.torrent_id, ['download_location']
            ).addCallback(self._on_show_file)

    def _set_file_priorities_on_user_change(self, selected, priority):
        """Sets the file priorities in the core. It will change the selected with the 'priority'"""
        file_priorities = []

        def set_file_priority(model, path, _iter, data):
            index = model.get_value(_iter, 5)
            if index in selected and index != -1:
                file_priorities.append((index, priority))
            elif index != -1:
                file_priorities.append((index, model.get_value(_iter, 4)))

        self.treestore.foreach(set_file_priority, None)
        file_priorities.sort()
        priorities = [p[1] for p in file_priorities]
        log.debug('priorities: %s', priorities)
        client.core.set_torrent_options(
            [self.torrent_id], {'file_priorities': priorities}
        )

    def on_menuitem_skip_activate(self, menuitem):
        self._set_file_priorities_on_user_change(
            self.get_selected_files(), FILE_PRIORITY['Skip']
        )

    def on_menuitem_low_activate(self, menuitem):
        self._set_file_priorities_on_user_change(
            self.get_selected_files(), FILE_PRIORITY['Low']
        )

    def on_menuitem_normal_activate(self, menuitem):
        self._set_file_priorities_on_user_change(
            self.get_selected_files(), FILE_PRIORITY['Normal']
        )

    def on_menuitem_high_activate(self, menuitem):
        self._set_file_priorities_on_user_change(
            self.get_selected_files(), FILE_PRIORITY['High']
        )

    def on_menuitem_expand_all_activate(self, menuitem):
        self.listview.expand_all()

    def _on_filename_edited(self, renderer, path, new_text):
        index = self.treestore[path][5]
        log.debug('new_text: %s', new_text)

        # Don't do anything if the text hasn't changed
        if new_text == self.treestore[path][0]:
            self._editing_index = None
            return

        if index > -1:
            # We are renaming a file
            itr = self.treestore.get_iter(path)
            # Recurse through the treestore to get the actual path of the file

            def get_filepath(i):
                ip = self.treestore.iter_parent(i)
                fp = ''
                while ip:
                    fp = self.treestore[ip][0] + fp
                    ip = self.treestore.iter_parent(ip)
                return fp

            # Only recurse if file is in a folder..
            if self.treestore.iter_parent(itr):
                filepath = get_filepath(itr) + new_text
            else:
                filepath = new_text

            log.debug('filepath: %s', filepath)

            client.core.rename_files(self.torrent_id, [(index, filepath)])
        else:
            # We are renaming a folder
            folder = self.treestore[path][0]

            parent_path = ''
            itr = self.treestore.iter_parent(self.treestore.get_iter(path))
            while itr:
                parent_path = self.treestore[itr][0] + parent_path
                itr = self.treestore.iter_parent(itr)

            client.core.rename_folder(
                self.torrent_id, parent_path + folder, parent_path + new_text
            )

        self._editing_index = None

    def _on_filename_editing_start(self, renderer, editable, path):
        self._editing_index = self.treestore[path][5]

    def _on_filename_editing_canceled(self, renderer):
        self._editing_index = None

    def _on_torrentfilerenamed_event(self, torrent_id, index, name):
        log.debug('index: %s name: %s', index, name)

        if torrent_id not in self.files_list:
            return

        old_name = self.files_list[torrent_id][index]['path']
        self.files_list[torrent_id][index]['path'] = name

        # We need to update the filename displayed if we're currently viewing
        # this torrents files.
        if torrent_id != self.torrent_id:
            return

        old_name_parent = old_name.split('/')[:-1]
        parent_path = name.split('/')[:-1]

        if old_name_parent != parent_path:
            if parent_path:
                for i, p in enumerate(parent_path):
                    p_itr = self.get_iter_at_path('/'.join(parent_path[: i + 1]) + '/')
                    if not p_itr:
                        p_itr = self.get_iter_at_path('/'.join(parent_path[:i]) + '/')
                        p_itr = self.treestore.append(
                            p_itr,
                            [parent_path[i] + '/', 0, '', 0, 0, -1, G_ICON_DIRECTORY],
                        )
                p_itr = self.get_iter_at_path('/'.join(parent_path) + '/')
                old_name_itr = self.get_iter_at_path(old_name)
                self.treestore.append(
                    p_itr,
                    self.treestore.get(
                        old_name_itr, *range(self.treestore.get_n_columns())
                    ),
                )
                self.treestore.remove(old_name_itr)

                # Remove old parent path
                p_itr = self.get_iter_at_path('/'.join(old_name_parent) + '/')
                self.remove_childless_folders(p_itr)
            else:
                new_folders = name.split('/')[:-1]
                parent_iter = None
                for f in new_folders:
                    parent_iter = self.treestore.append(
                        parent_iter, [f + '/', 0, '', 0, 0, -1, G_ICON_DIRECTORY]
                    )
                child = self.get_iter_at_path(old_name)
                self.treestore.append(
                    parent_iter,
                    self.treestore.get(child, *range(self.treestore.get_n_columns())),
                )
                self.treestore.remove(child)

        else:
            # This is just changing a filename without any folder changes
            def set_file_name(model, path, itr, user_data):
                if model[itr][5] == index:
                    model[itr][0] = os.path.split(name)[-1]
                    return True

            self.treestore.foreach(set_file_name, None)

    def get_iter_at_path(self, filepath):
        """Returns the gtkTreeIter for filepath."""
        log.debug('get_iter_at_path: %s', filepath)
        is_dir = False
        if filepath[-1] == '/':
            is_dir = True

        filepath = filepath.split('/')
        if '' in filepath:
            filepath.remove('')

        path_iter = None
        itr = self.treestore.iter_children(None)
        level = 0
        while itr:
            ipath = self.treestore[itr][0]
            if (level + 1) != len(filepath) and ipath == filepath[level] + '/':
                # We're not at the last index, but we do have a match
                itr = self.treestore.iter_children(itr)
                level += 1
                continue
            elif (level + 1) == len(filepath) and ipath == (
                filepath[level] + '/' if is_dir else filepath[level]
            ):
                # This is the iter we've been searching for
                path_iter = itr
                break
            else:
                itr = self.treestore.iter_next(itr)
                continue
        return path_iter

    def remove_childless_folders(self, itr):
        """Goes up the tree removing childless itrs starting at itr."""
        while not self.treestore.iter_children(itr):
            parent = self.treestore.iter_parent(itr)
            self.treestore.remove(itr)
            itr = parent

    def _on_torrentfolderrenamed_event(self, torrent_id, old_folder, new_folder):
        log.debug('on_torrent_folder_renamed_signal')
        log.debug('old_folder: %s new_folder: %s', old_folder, new_folder)

        if torrent_id not in self.files_list:
            return

        if old_folder[-1] != '/':
            old_folder += '/'

        if len(new_folder) > 0 and new_folder[-1] != '/':
            new_folder += '/'

        for fd in self.files_list[torrent_id]:
            if fd['path'].startswith(old_folder):
                fd['path'] = fd['path'].replace(old_folder, new_folder, 1)

        if torrent_id == self.torrent_id:
            old_split = old_folder.split('/')
            try:
                old_split.remove('')
            except ValueError:
                pass

            new_split = new_folder.split('/')
            try:
                new_split.remove('')
            except ValueError:
                pass

            old_folder_iter = self.get_iter_at_path(old_folder)
            old_folder_iter_parent = self.treestore.iter_parent(old_folder_iter)

            new_folder_iter = self.get_iter_at_path(new_folder) if new_folder else None

            if len(new_split) == len(old_split):
                # These are at the same tree depth, so it's a simple rename
                self.treestore[old_folder_iter][0] = new_split[-1] + '/'
                return
            if new_folder_iter:
                # This means that a folder by this name already exists
                reparent_iter(
                    self.treestore,
                    self.treestore.iter_children(old_folder_iter),
                    new_folder_iter,
                )
            else:
                parent = old_folder_iter_parent
                if new_split:
                    for ns in new_split[:-1]:
                        parent = self.treestore.append(
                            parent, [ns + '/', 0, '', 0, 0, -1, G_ICON_DIRECTORY]
                        )

                    self.treestore[old_folder_iter][0] = new_split[-1] + '/'
                    reparent_iter(self.treestore, old_folder_iter, parent)
                else:
                    child_itr = self.treestore.iter_children(old_folder_iter)
                    reparent_iter(
                        self.treestore,
                        child_itr,
                        old_folder_iter_parent,
                        move_siblings=True,
                    )

            # We need to check if the old_folder_iter no longer has children
            # and if so, we delete it
            self.remove_childless_folders(old_folder_iter)

    def _on_torrentremoved_event(self, torrent_id):
        if torrent_id in self.files_list:
            del self.files_list[torrent_id]

    def _on_drag_data_get_data(self, treeview, context, selection, target_id, etime):
        paths = self.listview.get_selection().get_selected_rows()[1]
        selection.set_text(json.dumps([str(path) for path in paths]), -1)

    def _on_drag_data_received_data(
        self, treeview, context, x, y, selection, info, etime
    ):
        try:
            selected = json.loads(selection.get_data())
        except TypeError:
            log.debug('Invalid selection data: %s', selection.get_data())
            return
        log.debug('selection.data: %s', selected)
        drop_info = treeview.get_dest_row_at_pos(x, y)
        model = treeview.get_model()
        if drop_info:
            itr = model.get_iter(drop_info[0])
            parent_iter = model.iter_parent(itr)
            parent_path = ''
            if model[itr][5] == -1:
                parent_path += model[itr][0]

            while parent_iter:
                parent_path = model[parent_iter][0] + parent_path
                parent_iter = model.iter_parent(parent_iter)

            if model[selected[0]][5] == -1:
                log.debug('parent_path: %s', parent_path)
                log.debug('rename_to: %s', parent_path + model[selected[0]][0])
                # Get the full path of the folder we want to rename
                pp = ''
                itr = self.treestore.iter_parent(self.treestore.get_iter(selected[0]))
                while itr:
                    pp = self.treestore[itr][0] + pp
                    itr = self.treestore.iter_parent(itr)
                client.core.rename_folder(
                    self.torrent_id,
                    pp + model[selected[0]][0],
                    parent_path + model[selected[0]][0],
                )
            else:
                # [(index, filepath), ...]
                to_rename = []
                for s in selected:
                    to_rename.append((model[s][5], parent_path + model[s][0]))
                log.debug('to_rename: %s', to_rename)
                client.core.rename_files(self.torrent_id, to_rename)



#
# Copyright (C) 2008 Martijn Voncken <mvoncken@gmail.com>
#               2008 Andrew Resch <andrewresch@gmail.com>
#               2014 Calum Lind <calumlind@gmail.com>
#
# This file is part of Deluge and is licensed under GNU General Public License 3.0, or later, with
# the additional special exception to link portions of this program with the OpenSSL library.
# See LICENSE for more details.
#

import logging
import os
import warnings

from gi.repository import Gtk
from gi.repository.GdkPixbuf import Pixbuf
from gi.repository.Pango import EllipsizeMode

import deluge.component as component
from deluge.common import TORRENT_STATE, decode_bytes, resource_filename
from deluge.configmanager import ConfigManager
from deluge.ui.client import client

from .common import get_pixbuf

log = logging.getLogger(__name__)

STATE_PIX = {
    'All': 'all',
    'Downloading': 'downloading',
    'Seeding': 'seeding',
    'Paused': 'inactive',
    'Checking': 'checking',
    'Queued': 'queued',
    'Error': 'alert',
    'Active': 'active',
    'Allocating': 'checking',
    'Moving': 'checking',
}

TRACKER_PIX = {'All': 'tracker_all', 'Error': 'tracker_warning'}

FILTER_COLUMN = 5


class FilterTreeView(component.Component):
    def __init__(self):
        component.Component.__init__(self, 'FilterTreeView', interval=2)
        self.config = ConfigManager('gtk3ui.conf')

        self.tracker_icons = component.get('TrackerIcons')

        self.sidebar = component.get('SideBar')
        self.treeview = Gtk.TreeView()
        self.sidebar.add_tab(self.treeview, 'filters', 'Filters')

        # set filter to all when hidden:
        self.sidebar.notebook.connect('hide', self._on_hide)

        # Create the treestore
        # cat, value, label, count, pixmap, visible
        self.treestore = Gtk.TreeStore(str, str, str, int, Pixbuf, bool)

        # Create the column and cells
        column = Gtk.TreeViewColumn('Filters')
        column.set_sizing(Gtk.TreeViewColumnSizing.AUTOSIZE)
        # icon cell
        self.cell_pix = Gtk.CellRendererPixbuf()
        column.pack_start(self.cell_pix, expand=False)
        column.add_attribute(self.cell_pix, 'pixbuf', 4)
        # label cell
        cell_label = Gtk.CellRendererText()
        cell_label.set_property('ellipsize', EllipsizeMode.END)
        column.pack_start(cell_label, expand=True)
        column.set_cell_data_func(cell_label, self.render_cell_data, None)
        # count cell
        self.cell_count = Gtk.CellRendererText()
        self.cell_count.set_property('xalign', 1.0)
        self.cell_count.set_padding(3, 0)
        column.pack_start(self.cell_count, expand=False)

        self.treeview.append_column(column)

        # Style
        self.treeview.set_show_expanders(True)
        self.treeview.set_headers_visible(False)
        self.treeview.set_level_indentation(-21)
        # Force theme to use expander-size so we don't cut out entries due to indentation hack.
        provider = Gtk.CssProvider()
        provider.load_from_data(b'* {-GtkTreeView-expander-size: 9;}')
        context = self.treeview.get_style_context()
        context.add_provider(provider, Gtk.STYLE_PROVIDER_PRIORITY_APPLICATION)

        self.treeview.set_model(self.treestore)
        self.treeview.get_selection().connect('changed', self.on_selection_changed)
        self.create_model_filter()

        self.treeview.connect('button-press-event', self.on_button_press_event)

        # filtertree menu
        builder = Gtk.Builder()
        builder.add_from_file(
            resource_filename(__package__, os.path.join('glade', 'filtertree_menu.ui'))
        )
        self.menu = builder.get_object('filtertree_menu')
        builder.connect_signals(self)

        self.default_menu_items = self.menu.get_children()

        # add Cat nodes:
        self.cat_nodes = {}
        self.filters = {}

    def start(self):
        self.cat_nodes = {}
        self.filters = {}
        # initial order of state filter:
        self.cat_nodes['state'] = self.treestore.append(
            None, ['cat', 'state', _('States'), 0, None, False]
        )
        for state in ['All', 'Active'] + TORRENT_STATE:
            self.update_row('state', state, 0, _(state))

        self.cat_nodes['tracker_host'] = self.treestore.append(
            None, ['cat', 'tracker_host', _('Trackers'), 0, None, False]
        )
        self.update_row('tracker_host', 'All', 0, _('All'))
        self.update_row('tracker_host', 'Error', 0, _('Error'))
        self.update_row('tracker_host', '', 0, _('None'))

        self.cat_nodes['owner'] = self.treestore.append(
            None, ['cat', 'owner', _('Owner'), 0, None, False]
        )
        self.update_row('owner', 'localclient', 0, _('Admin'))
        self.update_row('owner', '', 0, _('None'))

        # We set to this expand the rows on start-up
        self.expand_rows = True

        self.selected_path = None

    def stop(self):
        self.treestore.clear()

    def create_model_filter(self):
        self.model_filter = self.treestore.filter_new()
        self.model_filter.set_visible_column(FILTER_COLUMN)
        self.treeview.set_model(self.model_filter)

    def cb_update_filter_tree(self, filter_items):
        # create missing cat_nodes
        for cat in filter_items:
            if cat not in self.cat_nodes:
                label = _(cat)
                if cat == 'label':
                    label = _('Labels')
                self.cat_nodes[cat] = self.treestore.append(
                    None, ['cat', cat, label, 0, None, False]
                )

        # update rows
        visible_filters = []
        for cat, filters in filter_items.items():
            for value, count in filters:
                self.update_row(cat, value, count)
                visible_filters.append((cat, value))

        # hide root-categories not returned by core-part of the plugin.
        for cat in self.cat_nodes:
            self.treestore.set_value(
                self.cat_nodes[cat],
                FILTER_COLUMN,
                True if cat in filter_items else False,
            )

        # hide items not returned by core-plugin.
        for f in self.filters:
            if f not in visible_filters:
                self.treestore.set_value(self.filters[f], FILTER_COLUMN, False)

        if self.expand_rows:
            self.treeview.expand_all()
            self.expand_rows = False

        if not self.selected_path:
            self.select_default_filter()

    def update_row(self, cat, value, count, label=None):
        def on_get_icon(icon):
            if icon:
                self.set_row_image(cat, value, icon.get_filename())

        if (cat, value) in self.filters:
            row = self.filters[(cat, value)]
            self.treestore.set_value(row, 3, count)
        else:
            pix = self.get_pixmap(cat, value)

            if value == '':
                if cat == 'label':
                    label = _('No Label')
                elif cat == 'owner':
                    label = _('No Owner')
            elif not label and value:
                label = _(value)

            row = self.treestore.append(
                self.cat_nodes[cat], [cat, value, label, count, pix, True]
            )
            self.filters[(cat, value)] = row

            if cat == 'tracker_host' and value not in ('All', 'Error') and value:
                d = self.tracker_icons.fetch(value)
                d.addCallback(on_get_icon)

        self.treestore.set_value(row, FILTER_COLUMN, True)
        return row

    def render_cell_data(self, column, cell, model, row, data):
        cat = model.get_value(row, 0)
        label = decode_bytes(model.get_value(row, 2))
        count = model.get_value(row, 3)

        # Suppress Warning: g_object_set_qdata: assertion `G_IS_OBJECT (object)' failed
        original_filters = warnings.filters[:]
        warnings.simplefilter('ignore')
        try:
            pix = model.get_value(row, 4)
        finally:
            warnings.filters = original_filters

        self.cell_pix.set_property('visible', True if pix else False)

        if cat == 'cat':
            self.cell_count.set_property('visible', False)
            cell.set_padding(10, 2)
            label = '<b>%s</b>' % label
        else:
            count_txt = '<small>%s</small>' % count
            self.cell_count.set_property('markup', count_txt)
            self.cell_count.set_property('visible', True)
            cell.set_padding(2, 1)
        cell.set_property('markup', label)

    def get_pixmap(self, cat, value):
        pix = None
        if cat == 'state':
            pix = STATE_PIX.get(value, None)
        elif cat == 'tracker_host':
            pix = TRACKER_PIX.get(value, None)

        if pix:
            return get_pixbuf('%s16.png' % pix)

    def set_row_image(self, cat, value, filename):
        pix = get_pixbuf(filename, size=16)
        row = self.filters[(cat, value)]
        self.treestore.set_value(row, 4, pix)
        return False

    def on_selection_changed(self, selection):
        try:
            (model, row) = self.treeview.get_selection().get_selected()
            if not row:
                log.debug('nothing selected')
                return

            cat = model.get_value(row, 0)
            value = model.get_value(row, 1)

            filter_dict = {cat: [value]}
            if value == 'All' or cat == 'cat':
                filter_dict = {}

            component.get('TorrentView').set_filter(filter_dict)

            self.selected_path = model.get_path(row)

        except Exception as ex:
            log.debug(ex)
            # paths is likely None .. so lets return None
            return None

    def update(self):
        try:
            hide_cat = []
            if not self.config['sidebar_show_trackers']:
                hide_cat.append('tracker_host')
            if not self.config['sidebar_show_owners']:
                hide_cat.append('owner')
            client.core.get_filter_tree(
                self.config['sidebar_show_zero'], hide_cat
            ).addCallback(self.cb_update_filter_tree)
        except Exception as ex:
            log.debug(ex)

    # Callbacks #
    def on_button_press_event(self, widget, event):
        """This is a callback for showing the right-click context menu."""
        x, y = event.get_coords()
        path = self.treeview.get_path_at_pos(int(x), int(y))
        if not path:
            return
        path = path[0]
        cat = self.model_filter[path][0]

        if event.button == 1:
            # Prevent selecting a category label
            if cat == 'cat':
                if self.treeview.row_expanded(path):
                    self.treeview.collapse_row(path)
                else:
                    self.treeview.expand_row(path, False)
                    if not self.selected_path:
                        self.select_default_filter()
                    else:
                        self.treeview.get_selection().select_path(self.selected_path)
                return True

        elif event.button == 3:
            # assign current cat, value to self:
            x, y = event.get_coords()
            path = self.treeview.get_path_at_pos(int(x), int(y))
            if not path:
                return
            row = self.model_filter.get_iter(path[0])
            self.cat = self.model_filter.get_value(row, 0)
            self.value = self.model_filter.get_value(row, 1)
            self.count = self.model_filter.get_value(row, 3)

            # Show the pop-up menu
            self.set_menu_sensitivity()
            self.menu.hide()
            self.menu.popup(None, None, None, None, event.button, event.time)
            self.menu.show()

            if cat == 'cat':
                # Do not select the row
                return True

    def set_menu_sensitivity(self):
        # select-all/pause/resume
        sensitive = self.cat != 'cat' and self.count != 0
        for item in self.default_menu_items:
            item.set_sensitive(sensitive)

    def select_all(self):
        """For use in popup menu."""
        component.get('TorrentView').treeview.get_selection().select_all()

    def on_select_all(self, event):
        self.select_all()

    def on_pause_all(self, event):
        self.select_all()
        func = getattr(component.get('MenuBar'), 'on_menuitem_%s_activate' % 'pause')
        func(event)

    def on_resume_all(self, event):
        self.select_all()
        func = getattr(component.get('MenuBar'), 'on_menuitem_%s_activate' % 'resume')
        func(event)

    def _on_hide(self, *args):
        self.select_default_filter()

    def select_default_filter(self):
        row = self.filters[('state', 'All')]
        path = self.treestore.get_path(row)
        self.treeview.get_selection().select_path(path)



#
# Copyright (C) 2007, 2008 Andrew Resch <andrewresch@gmail.com>
# Copyright (C) 2011 Pedro Algarvio <pedro@algarvio.me>
#
# This file is part of Deluge and is licensed under GNU General Public License 3.0, or later, with
# the additional special exception to link portions of this program with the OpenSSL library.
# See LICENSE for more details.
#


import logging
import os.path

from gi.repository import Gtk

import deluge.common
import deluge.component as component
from deluge.configmanager import ConfigManager
from deluge.ui.client import client

from .dialogs import CopyMagnetDialog, ErrorDialog, OtherDialog
from .path_chooser import PathChooser

log = logging.getLogger(__name__)

default_main_window_accelmap = {
    '<Deluge-MainWindow>/File/Add Torrent': '<Primary>o',
    '<Deluge-MainWindow>/File/Create Torrent': '<Primary>n',
    '<Deluge-MainWindow>/File/Quit & Shutdown Daemon': '<Primary><Shift>q',
    '<Deluge-MainWindow>/File/Quit': '<Primary>q',
    '<Deluge-MainWindow>/Edit/Preferences': '<Primary>p',
    '<Deluge-MainWindow>/Edit/Connection Manager': '<Primary>m',
    '<Deluge-MainWindow>/View/Find ...': '<Primary>f',
    '<Deluge-MainWindow>/Help/FAQ': 'F1',
}


class MenuBar(component.Component):
    def __init__(self):
        log.debug('MenuBar init..')
        component.Component.__init__(self, 'MenuBar')
        self.mainwindow = component.get('MainWindow')
        self.main_builder = self.mainwindow.get_builder()
        self.config = ConfigManager('gtk3ui.conf')
        self._magnet_copied = False

        self.builder = Gtk.Builder()
        # Get the torrent menu from the gtk builder file
        self.builder.add_from_file(
            deluge.common.resource_filename(
                __package__, os.path.join('glade', 'torrent_menu.ui')
            )
        )
        # Get the torrent options menu from the gtk builder file
        self.builder.add_from_file(
            deluge.common.resource_filename(
                __package__, os.path.join('glade', 'torrent_menu.options.ui')
            )
        )
        # Get the torrent queue menu from the gtk builder file
        self.builder.add_from_file(
            deluge.common.resource_filename(
                __package__, os.path.join('glade', 'torrent_menu.queue.ui')
            )
        )

        # Attach queue torrent menu
        torrent_queue_menu = self.builder.get_object('queue_torrent_menu')
        self.builder.get_object('menuitem_queue').set_submenu(torrent_queue_menu)
        # Attach options torrent menu
        torrent_options_menu = self.builder.get_object('options_torrent_menu')
        self.builder.get_object('menuitem_options').set_submenu(torrent_options_menu)

        self.builder.get_object('download-limit-image').set_from_file(
            deluge.common.get_pixmap('downloading16.png')
        )
        self.builder.get_object('upload-limit-image').set_from_file(
            deluge.common.get_pixmap('seeding16.png')
        )

        for menuitem in (
            'menuitem_down_speed',
            'menuitem_up_speed',
            'menuitem_max_connections',
            'menuitem_upload_slots',
        ):
            submenu = Gtk.Menu()
            item = Gtk.MenuItem.new_with_label(_('Set Unlimited'))
            item.set_name(menuitem)
            item.connect('activate', self.on_menuitem_set_unlimited)
            submenu.append(item)
            item = Gtk.MenuItem.new_with_label(_('Other...'))
            item.set_name(menuitem)
            item.connect('activate', self.on_menuitem_set_other)
            submenu.append(item)
            submenu.show_all()
            self.builder.get_object(menuitem).set_submenu(submenu)

        submenu = Gtk.Menu()
        item = Gtk.MenuItem.new_with_label(_('On'))
        item.connect('activate', self.on_menuitem_set_automanaged_on)
        submenu.append(item)
        item = Gtk.MenuItem.new_with_label(_('Off'))
        item.connect('activate', self.on_menuitem_set_automanaged_off)
        submenu.append(item)
        submenu.show_all()
        self.builder.get_object('menuitem_auto_managed').set_submenu(submenu)

        submenu = Gtk.Menu()
        item = Gtk.MenuItem.new_with_label(_('Disable'))
        item.connect('activate', self.on_menuitem_set_stop_seed_at_ratio_disable)
        submenu.append(item)
        item = Gtk.MenuItem.new_with_label(_('Enable...'))
        item.set_name('menuitem_stop_seed_at_ratio')
        item.connect('activate', self.on_menuitem_set_other)
        submenu.append(item)
        submenu.show_all()
        self.builder.get_object('menuitem_stop_seed_at_ratio').set_submenu(submenu)

        self.torrentmenu = self.builder.get_object('torrent_menu')
        self.menu_torrent = self.main_builder.get_object('menu_torrent')

        # Attach the torrent_menu to the Torrent file menu
        self.menu_torrent.set_submenu(self.torrentmenu)

        # Set keyboard shortcuts
        for accel_path, accelerator in default_main_window_accelmap.items():
            accel_key, accel_mods = Gtk.accelerator_parse(accelerator)
            Gtk.AccelMap.change_entry(accel_path, accel_key, accel_mods, True)

        # Make sure the view menuitems are showing the correct active state
        self.main_builder.get_object('menuitem_toolbar').set_active(
            self.config['show_toolbar']
        )
        self.main_builder.get_object('menuitem_sidebar').set_active(
            self.config['show_sidebar']
        )
        self.main_builder.get_object('menuitem_statusbar').set_active(
            self.config['show_statusbar']
        )
        self.main_builder.get_object('sidebar_show_zero').set_active(
            self.config['sidebar_show_zero']
        )
        self.main_builder.get_object('sidebar_show_trackers').set_active(
            self.config['sidebar_show_trackers']
        )
        self.main_builder.get_object('sidebar_show_owners').set_active(
            self.config['sidebar_show_owners']
        )

        # Connect main window Signals #
        self.mainwindow.connect_signals(self)

        # Connect menubar signals
        self.builder.connect_signals(self)

        self.change_sensitivity = ['menuitem_addtorrent']

    def magnet_copied(self):
        """
        lets the caller know whether a magnet was copied internally

        the `mainwindow` checks every time the data in the clipboard,
        so it will automatically open the AddTorrentURL dialog in case it
        contains a valid link (URL to a torrent or a magnet URI).

        """
        val = self._magnet_copied
        self._magnet_copied = False
        return val

    def start(self):
        for widget in self.change_sensitivity:
            self.main_builder.get_object(widget).set_sensitive(True)

        # Only show open_folder menuitem and separator if connected to a localhost daemon.
        localhost_items = ['menuitem_open_folder', 'separator4']
        if client.is_localhost():
            for widget in localhost_items:
                self.builder.get_object(widget).show()
                self.builder.get_object(widget).set_no_show_all(False)
        else:
            for widget in localhost_items:
                self.builder.get_object(widget).hide()
                self.builder.get_object(widget).set_no_show_all(True)

        self.main_builder.get_object('separatormenuitem').set_visible(
            not self.config['standalone']
        )
        self.main_builder.get_object('menuitem_quitdaemon').set_visible(
            not self.config['standalone']
        )
        self.main_builder.get_object('menuitem_connectionmanager').set_visible(
            not self.config['standalone']
        )

        # Show the Torrent menu because we're connected to a host
        self.menu_torrent.show()

        if client.get_auth_level() == deluge.common.AUTH_LEVEL_ADMIN:
            # Get known accounts to allow changing ownership
            client.core.get_known_accounts().addCallback(
                self._on_known_accounts
            ).addErrback(self._on_known_accounts_fail)

        client.register_event_handler(
            'TorrentStateChangedEvent', self.on_torrentstatechanged_event
        )
        client.register_event_handler(
            'TorrentResumedEvent', self.on_torrentresumed_event
        )
        client.register_event_handler('SessionPausedEvent', self.on_sessionpaused_event)
        client.register_event_handler(
            'SessionResumedEvent', self.on_sessionresumed_event
        )

    def stop(self):
        log.debug('MenuBar stopping')

        client.deregister_event_handler(
            'TorrentStateChangedEvent', self.on_torrentstatechanged_event
        )
        client.deregister_event_handler(
            'TorrentResumedEvent', self.on_torrentresumed_event
        )
        client.deregister_event_handler(
            'SessionPausedEvent', self.on_sessionpaused_event
        )
        client.deregister_event_handler(
            'SessionResumedEvent', self.on_sessionresumed_event
        )

        for widget in self.change_sensitivity:
            self.main_builder.get_object(widget).set_sensitive(False)

        # Hide the Torrent menu
        self.menu_torrent.hide()

        self.main_builder.get_object('separatormenuitem').hide()
        self.main_builder.get_object('menuitem_quitdaemon').hide()

    def update_menu(self):
        selected = component.get('TorrentView').get_selected_torrents()
        if not selected or len(selected) == 0:
            # No torrent is selected. Disable the 'Torrents' menu
            self.menu_torrent.set_sensitive(False)
            return

        self.menu_torrent.set_sensitive(True)
        # XXX: Should also update Pause/Resume/Remove menuitems.
        # Any better way than duplicating toolbar.py:update_buttons in here?

    def add_torrentmenu_separator(self):
        sep = Gtk.SeparatorMenuItem()
        self.torrentmenu.append(sep)
        sep.show()
        return sep

    # Callbacks #
    def on_torrentstatechanged_event(self, torrent_id, state):
        if state == 'Paused':
            self.update_menu()

    def on_torrentresumed_event(self, torrent_id):
        self.update_menu()

    def on_sessionpaused_event(self):
        self.update_menu()

    def on_sessionresumed_event(self):
        self.update_menu()

    # File Menu #
    def on_menuitem_addtorrent_activate(self, data=None):
        log.debug('on_menuitem_addtorrent_activate')
        component.get('AddTorrentDialog').show()

    def on_menuitem_createtorrent_activate(self, data=None):
        log.debug('on_menuitem_createtorrent_activate')
        from .createtorrentdialog import CreateTorrentDialog

        CreateTorrentDialog().show()

    def on_menuitem_quitdaemon_activate(self, data=None):
        log.debug('on_menuitem_quitdaemon_activate')
        self.mainwindow.quit(shutdown=True)

    def on_menuitem_quit_activate(self, data=None):
        log.debug('on_menuitem_quit_activate')
        self.mainwindow.quit()

    # Edit Menu #
    def on_menuitem_preferences_activate(self, data=None):
        log.debug('on_menuitem_preferences_activate')
        component.get('Preferences').show()

    def on_menuitem_connectionmanager_activate(self, data=None):
        log.debug('on_menuitem_connectionmanager_activate')
        component.get('ConnectionManager').show()

    # Torrent Menu #
    def on_menuitem_pause_activate(self, data=None):
        log.debug('on_menuitem_pause_activate')
        client.core.pause_torrents(component.get('TorrentView').get_selected_torrents())

    def on_menuitem_resume_activate(self, data=None):
        log.debug('on_menuitem_resume_activate')
        client.core.resume_torrents(
            component.get('TorrentView').get_selected_torrents()
        )

    def on_menuitem_copymagnet_activate(self, data=None):
        log.debug('on_menuitem_copymagnet_activate')
        torrent_ids = component.get('TorrentView').get_selected_torrents()
        if torrent_ids:

            def _on_magnet_uri(magnet_uri):
                def update_copied(response_id):
                    if dialog.copied:
                        self._magnet_copied = True

                dialog = CopyMagnetDialog(magnet_uri)
                dialog.run().addCallback(update_copied)

            client.core.get_magnet_uri(torrent_ids[0]).addCallback(_on_magnet_uri)

    def on_menuitem_updatetracker_activate(self, data=None):
        log.debug('on_menuitem_updatetracker_activate')
        client.core.force_reannounce(
            component.get('TorrentView').get_selected_torrents()
        )

    def on_menuitem_edittrackers_activate(self, data=None):
        log.debug('on_menuitem_edittrackers_activate')
        from .edittrackersdialog import EditTrackersDialog

        dialog = EditTrackersDialog(
            component.get('TorrentView').get_selected_torrent(), self.mainwindow.window
        )
        dialog.run()

    def on_menuitem_remove_activate(self, data=None):
        log.debug('on_menuitem_remove_activate')
        torrent_ids = component.get('TorrentView').get_selected_torrents()
        if torrent_ids:
            from .removetorrentdialog import RemoveTorrentDialog

            RemoveTorrentDialog(torrent_ids).run()

    def on_menuitem_recheck_activate(self, data=None):
        log.debug('on_menuitem_recheck_activate')
        client.core.force_recheck(component.get('TorrentView').get_selected_torrents())

    def on_menuitem_open_folder_activate(self, data=None):
        log.debug('on_menuitem_open_folder')

        def _on_torrent_status(status):
            timestamp = component.get('MainWindow').get_timestamp()
            path = os.path.join(
                status['download_location'], status['files'][0]['path'].split('/')[0]
            )
            deluge.common.show_file(path, timestamp=timestamp)

        for torrent_id in component.get('TorrentView').get_selected_torrents():
            component.get('SessionProxy').get_torrent_status(
                torrent_id, ['download_location', 'files']
            ).addCallback(_on_torrent_status)

    def on_menuitem_move_activate(self, data=None):
        log.debug('on_menuitem_move_activate')
        component.get('SessionProxy').get_torrent_status(
            component.get('TorrentView').get_selected_torrent(), ['download_location']
        ).addCallback(self.show_move_storage_dialog)

    def show_move_storage_dialog(self, status):
        log.debug('show_move_storage_dialog')
        builder = Gtk.Builder()
        builder.add_from_file(
            deluge.common.resource_filename(
                __package__, os.path.join('glade', 'move_storage_dialog.ui')
            )
        )
        # Keep it referenced:
        #  https://bugzilla.gnome.org/show_bug.cgi?id=546802
        self.move_storage_dialog = builder.get_object('move_storage_dialog')
        self.move_storage_dialog.set_transient_for(self.mainwindow.window)
        self.move_storage_dialog_hbox = builder.get_object('hbox_entry')
        self.move_storage_path_chooser = PathChooser(
            'move_completed_paths_list', self.move_storage_dialog
        )
        self.move_storage_dialog_hbox.add(self.move_storage_path_chooser)
        self.move_storage_dialog_hbox.show_all()
        self.move_storage_path_chooser.set_text(status['download_location'])

        def on_dialog_response_event(widget, response_id):
            def on_core_result(result):
                # Delete references
                self.move_storage_dialog.hide()
                del self.move_storage_dialog
                del self.move_storage_dialog_hbox

            if response_id == Gtk.ResponseType.CANCEL:
                on_core_result(None)

            if response_id == Gtk.ResponseType.OK:
                log.debug(
                    'Moving torrents to %s', self.move_storage_path_chooser.get_text()
                )
                path = self.move_storage_path_chooser.get_text()
                client.core.move_storage(
                    component.get('TorrentView').get_selected_torrents(), path
                ).addCallback(on_core_result)

        self.move_storage_dialog.connect('response', on_dialog_response_event)
        self.move_storage_dialog.show()

    def on_menuitem_queue_top_activate(self, value):
        log.debug('on_menuitem_queue_top_activate')
        client.core.queue_top(component.get('TorrentView').get_selected_torrents())

    def on_menuitem_queue_up_activate(self, value):
        log.debug('on_menuitem_queue_up_activate')
        client.core.queue_up(component.get('TorrentView').get_selected_torrents())

    def on_menuitem_queue_down_activate(self, value):
        log.debug('on_menuitem_queue_down_activate')
        client.core.queue_down(component.get('TorrentView').get_selected_torrents())

    def on_menuitem_queue_bottom_activate(self, value):
        log.debug('on_menuitem_queue_bottom_activate')
        client.core.queue_bottom(component.get('TorrentView').get_selected_torrents())

    # View Menu #
    def on_menuitem_toolbar_toggled(self, value):
        log.debug('on_menuitem_toolbar_toggled')
        component.get('ToolBar').visible(value.get_active())

    def on_menuitem_sidebar_toggled(self, value):
        log.debug('on_menuitem_sidebar_toggled')
        component.get('SideBar').visible(value.get_active())

    def on_menuitem_statusbar_toggled(self, value):
        log.debug('on_menuitem_statusbar_toggled')
        component.get('StatusBar').visible(value.get_active())

    # Help Menu #
    def on_menuitem_homepage_activate(self, data=None):
        log.debug('on_menuitem_homepage_activate')
        deluge.common.open_url_in_browser('http://deluge-torrent.org')

    def on_menuitem_faq_activate(self, data=None):
        log.debug('on_menuitem_faq_activate')
        deluge.common.open_url_in_browser('http://dev.deluge-torrent.org/wiki/Faq')

    def on_menuitem_community_activate(self, data=None):
        log.debug('on_menuitem_community_activate')
        deluge.common.open_url_in_browser('http://forum.deluge-torrent.org/')

    def on_menuitem_about_activate(self, data=None):
        log.debug('on_menuitem_about_activate')
        from .aboutdialog import AboutDialog

        AboutDialog().run()

    def on_menuitem_set_unlimited(self, widget):
        log.debug('widget name: %s', widget.get_name())
        funcs = {
            'menuitem_down_speed': 'max_download_speed',
            'menuitem_up_speed': 'max_upload_speed',
            'menuitem_max_connections': 'max_connections',
            'menuitem_upload_slots': 'max_upload_slots',
        }
        if widget.get_name() in funcs:
            torrent_ids = component.get('TorrentView').get_selected_torrents()
            client.core.set_torrent_options(torrent_ids, {funcs[widget.get_name()]: -1})

    def on_menuitem_set_other(self, widget):
        log.debug('widget name: %s', widget.get_name())
        status_map = {
            'menuitem_down_speed': ['max_download_speed', 'max_download_speed'],
            'menuitem_up_speed': ['max_upload_speed', 'max_upload_speed'],
            'menuitem_max_connections': ['max_connections', 'max_connections_global'],
            'menuitem_upload_slots': ['max_upload_slots', 'max_upload_slots_global'],
            'menuitem_stop_seed_at_ratio': ['stop_ratio', 'stop_seed_ratio'],
        }

        other_dialog_info = {
            'menuitem_down_speed': [
                _('Download Speed Limit'),
                _('Set the maximum download speed'),
                _('KiB/s'),
                'downloading.svg',
            ],
            'menuitem_up_speed': [
                _('Upload Speed Limit'),
                _('Set the maximum upload speed'),
                _('KiB/s'),
                'seeding.svg',
            ],
            'menuitem_max_connections': [
                _('Incoming Connections'),
                _('Set the maximum incoming connections'),
                '',
                'network-transmit-receive-symbolic',
            ],
            'menuitem_upload_slots': [
                _('Peer Upload Slots'),
                _('Set the maximum upload slots'),
                '',
                'view-sort-descending-symbolic',
            ],
            'menuitem_stop_seed_at_ratio': [
                _('Stop Seed At Ratio'),
                'Stop torrent seeding at ratio',
                '',
                None,
            ],
        }

        core_key = status_map[widget.get_name()][0]
        core_key_global = status_map[widget.get_name()][1]

        def _on_torrent_status(status):
            other_dialog = other_dialog_info[widget.get_name()]
            # Add the default using status value
            if status:
                other_dialog.append(status[core_key_global])

            def set_value(value):
                if value is not None:
                    if value == 0:
                        value += -1
                    options = {core_key: value}
                    if core_key == 'stop_ratio':
                        options['stop_at_ratio'] = True
                    client.core.set_torrent_options(torrent_ids, options)

            dialog = OtherDialog(*other_dialog)
            dialog.run().addCallback(set_value)

        torrent_ids = component.get('TorrentView').get_selected_torrents()
        if len(torrent_ids) == 1:
            core_key_global = core_key
            d = component.get('SessionProxy').get_torrent_status(
                torrent_ids[0], [core_key]
            )
        else:
            d = client.core.get_config_values([core_key_global])
        d.addCallback(_on_torrent_status)

    def on_menuitem_set_automanaged_on(self, widget):
        client.core.set_torrent_options(
            component.get('TorrentView').get_selected_torrents(), {'auto_managed': True}
        )

    def on_menuitem_set_automanaged_off(self, widget):
        client.core.set_torrent_options(
            component.get('TorrentView').get_selected_torrents(),
            {'auto_managed': False},
        )

    def on_menuitem_set_stop_seed_at_ratio_disable(self, widget):
        client.core.set_torrent_options(
            component.get('TorrentView').get_selected_torrents(),
            {'stop_at_ratio': False},
        )

    def on_menuitem_sidebar_zero_toggled(self, widget):
        self.config['sidebar_show_zero'] = widget.get_active()
        component.get('FilterTreeView').update()

    def on_menuitem_sidebar_trackers_toggled(self, widget):
        self.config['sidebar_show_trackers'] = widget.get_active()
        component.get('FilterTreeView').update()

    def on_menuitem_sidebar_owners_toggled(self, widget):
        self.config['sidebar_show_owners'] = widget.get_active()
        component.get('FilterTreeView').update()

    def _on_known_accounts(self, known_accounts):
        menuitem_change_owner = self.builder.get_object('menuitem_change_owner')
        if len(known_accounts) <= 1:
            menuitem_change_owner.set_visible(False)
            return

        self.users_menu = Gtk.Menu()
        self.users_menu_items = {}
        menu_group = None

        for account in known_accounts:
            username = account['username']
            item = Gtk.RadioMenuItem.new_with_label(menu_group, username)
            menu_group = item.get_group()
            item.connect('toggled', self._on_change_owner_toggled, username)
            self.users_menu_items[username] = item
            self.users_menu.append(item)

        self.users_menu.show_all()
        menuitem_change_owner.set_submenu(self.users_menu)
        menuitem_change_owner.set_visible(True)

    def _on_known_accounts_fail(self, reason):
        self.builder.get_object('menuitem_change_owner').set_visible(False)

    def _on_change_owner_submenu_active(self, widget):
        log.debug('_on_change_owner_submenu_active')
        selected = component.get('TorrentView').get_selected_torrents()
        if len(selected) > 1:
            self.users_menu_items[None].set_active(True)
            return

        torrent_owner = component.get('TorrentView').get_torrent_status(selected[0])[
            'owner'
        ]
        for username, item in self.users_menu_items.items():
            item.set_active(username == torrent_owner)

    def _on_change_owner_toggled(self, widget, username):
        log.debug('_on_change_owner_toggled')
        update_torrents = []
        selected = component.get('TorrentView').get_selected_torrents()
        for torrent_id in selected:
            torrent_status = component.get('TorrentView').get_torrent_status(torrent_id)
            if torrent_status['owner'] != username:
                update_torrents.append(torrent_id)

        if update_torrents:
            log.debug('Setting torrent owner "%s" on %s', username, update_torrents)

            def failed_change_owner(failure):
                ErrorDialog(
                    _('Ownership Change Error'),
                    _('There was an error while trying changing ownership.'),
                    self.mainwindow.window,
                    details=failure.value.logable(),
                ).run()

            client.core.set_torrent_options(
                update_torrents, {'owner': username}
            ).addErrback(failed_change_owner)



#
# Copyright (C) 2008 Andrew Resch <andrewresch@gmail.com>
#               2017 Calum Lind <calumlind+deluge@gmail.com>
#
# This file is part of Deluge and is licensed under GNU General Public License 3.0, or later, with
# the additional special exception to link portions of this program with the OpenSSL library.
# See LICENSE for more details.
#

from gi.repository.Gdk import keyval_name

import deluge.component as component
from deluge.ui.client import client

from .path_chooser import PathChooser
from .torrentdetails import Tab


class OptionsTab(Tab):
    def __init__(self):
        super().__init__('Options', 'options_tab', 'options_tab_label')

        self.prev_torrent_ids = None
        self.prev_status = None
        self.inconsistent_keys = []

        # Create TabWidget items with widget id, get/set func name, status key.
        self.add_tab_widget('spin_max_download', 'value', ['max_download_speed'])
        self.add_tab_widget('spin_max_upload', 'value', ['max_upload_speed'])
        self.add_tab_widget('spin_max_connections', 'value_as_int', ['max_connections'])
        self.add_tab_widget(
            'spin_max_upload_slots', 'value_as_int', ['max_upload_slots']
        )
        self.add_tab_widget(
            'chk_prioritize_first_last', 'active', ['prioritize_first_last_pieces']
        )
        self.add_tab_widget(
            'chk_sequential_download', 'active', ['sequential_download']
        )
        self.add_tab_widget('chk_auto_managed', 'active', ['auto_managed'])
        self.add_tab_widget('chk_stop_at_ratio', 'active', ['stop_at_ratio'])
        self.add_tab_widget('chk_remove_at_ratio', 'active', ['remove_at_ratio'])
        self.add_tab_widget('spin_stop_ratio', 'value', ['stop_ratio'])
        self.add_tab_widget('chk_move_completed', 'active', ['move_completed'])
        self.add_tab_widget('chk_shared', 'active', ['shared'])
        self.add_tab_widget('summary_owner', 'text', ['owner'])
        self.add_tab_widget('chk_super_seeding', 'active', ['super_seeding'])

        # Connect key press event for spin widgets.
        for widget_id in self.tab_widgets:
            if widget_id.startswith('spin_'):
                self.tab_widgets[widget_id].obj.connect(
                    'key-press-event', self.on_key_press_event
                )

        self.button_apply = self.main_builder.get_object('button_apply')

        self.move_completed_path_chooser = PathChooser(
            'move_completed_paths_list', parent=component.get('MainWindow').window
        )
        self.move_completed_path_chooser.set_sensitive(
            self.tab_widgets['chk_move_completed'].obj.get_active()
        )
        self.move_completed_path_chooser.connect(
            'text-changed', self.on_path_chooser_text_changed_event
        )
        self.status_keys.append('move_completed_path')

        self.move_completed_hbox = self.main_builder.get_object(
            'hbox_move_completed_path_chooser'
        )
        self.move_completed_hbox.add(self.move_completed_path_chooser)
        self.move_completed_hbox.show_all()

        component.get('MainWindow').connect_signals(self)

    def start(self):
        pass

    def stop(self):
        pass

    def clear(self):
        self.prev_torrent_ids = None
        self.prev_status = None
        self.inconsistent_keys = []

    def update(self):
        torrent_ids = component.get('TorrentView').get_selected_torrents()

        # Set True if torrent(s) selected in torrentview, else False.
        self._child_widget.set_sensitive(bool(torrent_ids))

        if torrent_ids:
            if torrent_ids != self.prev_torrent_ids:
                self.clear()

            component.get('SessionProxy').get_torrents_status(
                {'id': torrent_ids}, self.status_keys
            ).addCallback(self.parse_torrents_statuses)

            self.prev_torrent_ids = torrent_ids

    def parse_torrents_statuses(self, statuses):
        """Finds common status values to all torrrents in statuses.

        Values which differ are replaced with config values.


        Args:
            statuses (dict): A status dict of {torrent_id: {key: value}}.

        Returns:
            dict: A single status dict.

        """
        status = {}
        if len(statuses) == 1:
            # A single torrent so pop torrent status.
            status = statuses.popitem()[1]
            self.button_apply.set_label('_Apply')
        else:
            for status_key in self.status_keys:
                prev_value = None
                for idx, status in enumerate(statuses.values()):
                    if idx == 0:
                        prev_value = status[status_key]
                        continue
                    elif status[status_key] != prev_value:
                        self.inconsistent_keys.append(status_key)
                        break
                status[status_key] = prev_value
            self.button_apply.set_label(_('_Apply to selected'))

        self.on_get_torrent_status(status)

    def on_get_torrent_status(self, new_status):
        # So we don't overwrite the user's unapplied changes we only
        # want to update values that have been applied in the core.
        if self.prev_status is None:
            self.prev_status = dict.fromkeys(new_status, None)

        if new_status != self.prev_status:
            for widget in self.tab_widgets.values():
                status_key = widget.status_keys[0]
                status_value = new_status[status_key]
                if status_value != self.prev_status[status_key]:
                    set_func = 'set_' + widget.func.replace('_as_int', '')
                    getattr(widget.obj, set_func)(status_value)
                    if set_func == 'set_active':
                        widget.obj.set_inconsistent(
                            status_key in self.inconsistent_keys
                        )

            if (
                new_status['move_completed_path']
                != self.prev_status['move_completed_path']
            ):
                text = new_status['move_completed_path']
                self.move_completed_path_chooser.set_text(
                    text, cursor_end=False, default_text=True
                )

            # Update sensitivity of widgets.
            self.tab_widgets['spin_stop_ratio'].obj.set_sensitive(
                new_status['stop_at_ratio']
            )
            self.tab_widgets['chk_remove_at_ratio'].obj.set_sensitive(
                new_status['stop_at_ratio']
            )

            # Ensure apply button sensitivity is set False.
            self.button_apply.set_sensitive(False)
            self.prev_status = new_status

    # === Widget signal handlers === #

    def on_button_apply_clicked(self, button):
        options = {}
        for widget in self.tab_widgets.values():
            status_key = widget.status_keys[0]
            if status_key == 'owner':
                continue  # A label so read-only
            widget_value = getattr(widget.obj, 'get_' + widget.func)()
            if widget_value != self.prev_status[status_key] or (
                status_key in self.inconsistent_keys
                and not widget.obj.get_inconsistent()
            ):
                options[status_key] = widget_value

        move_completed_path = self.move_completed_path_chooser.get_text()
        if move_completed_path != self.prev_status['move_completed_path']:
            options['move_completed_path'] = move_completed_path

        client.core.set_torrent_options(self.prev_torrent_ids, options)
        self.button_apply.set_sensitive(False)

    def on_chk_move_completed_toggled(self, widget):
        self.move_completed_path_chooser.set_sensitive(widget.get_active())
        self.on_chk_toggled(widget)

    def on_chk_stop_at_ratio_toggled(self, widget):
        self.tab_widgets['spin_stop_ratio'].obj.set_sensitive(widget.get_active())
        self.tab_widgets['chk_remove_at_ratio'].obj.set_sensitive(widget.get_active())
        self.on_chk_toggled(widget)

    def on_chk_toggled(self, widget):
        widget.set_inconsistent(False)
        self.button_apply.set_sensitive(True)

    def on_spin_value_changed(self, widget):
        self.button_apply.set_sensitive(True)

    def on_key_press_event(self, widget, event):
        keyname = keyval_name(event.keyval).lstrip('KP_').lower()
        if keyname.isdigit() or keyname in ['period', 'minus', 'delete', 'backspace']:
            self.button_apply.set_sensitive(True)

    def on_path_chooser_text_changed_event(self, widget, path):
        self.button_apply.set_sensitive(True)



#
# Copyright (C) 2008 Andrew Resch <andrewresch@gmail.com>
#
# This file is part of Deluge and is licensed under GNU General Public License 3.0, or later, with
# the additional special exception to link portions of this program with the OpenSSL library.
# See LICENSE for more details.
#

import logging

import deluge.component as component
from deluge.common import decode_bytes, fpeer
from deluge.configmanager import ConfigManager

from .piecesbar import PiecesBar
from .tab_data_funcs import (
    fdate_or_never,
    fpcnt,
    fratio,
    fseed_rank_or_dash,
    fspeed_max,
    ftime_or_dash,
    ftotal_sized,
)
from .torrentdetails import Tab, TabWidget

log = logging.getLogger(__name__)


class StatusTab(Tab):
    def __init__(self):
        super().__init__('Status', 'status_tab', 'status_tab_label')

        self.config = ConfigManager('gtk3ui.conf')

        self.progressbar = self.main_builder.get_object('progressbar')
        self.piecesbar = None

        self.add_tab_widget('summary_availability', fratio, ('distributed_copies',))
        self.add_tab_widget(
            'summary_total_downloaded',
            ftotal_sized,
            ('all_time_download', 'total_payload_download'),
        )
        self.add_tab_widget(
            'summary_total_uploaded',
            ftotal_sized,
            ('total_uploaded', 'total_payload_upload'),
        )
        self.add_tab_widget(
            'summary_download_speed',
            fspeed_max,
            ('download_payload_rate', 'max_download_speed'),
        )
        self.add_tab_widget(
            'summary_upload_speed',
            fspeed_max,
            ('upload_payload_rate', 'max_upload_speed'),
        )
        self.add_tab_widget('summary_seeds', fpeer, ('num_seeds', 'total_seeds'))
        self.add_tab_widget('summary_peers', fpeer, ('num_peers', 'total_peers'))
        self.add_tab_widget('summary_eta', ftime_or_dash, ('eta',))
        self.add_tab_widget('summary_share_ratio', fratio, ('ratio',))
        self.add_tab_widget('summary_active_time', ftime_or_dash, ('active_time',))
        self.add_tab_widget('summary_seed_time', ftime_or_dash, ('seeding_time',))
        self.add_tab_widget(
            'summary_seed_rank', fseed_rank_or_dash, ('seed_rank', 'seeding_time')
        )
        self.add_tab_widget('progressbar', fpcnt, ('progress', 'state', 'message'))
        self.add_tab_widget(
            'summary_last_seen_complete', fdate_or_never, ('last_seen_complete',)
        )
        self.add_tab_widget(
            'summary_last_transfer', ftime_or_dash, ('time_since_transfer',)
        )

        self.config.register_set_function(
            'show_piecesbar', self.on_show_piecesbar_config_changed, apply_now=True
        )

    def update(self):
        # Get the first selected torrent
        selected = component.get('TorrentView').get_selected_torrent()

        if not selected:
            # No torrent is selected in the torrentview
            self.clear()
            return

        # Get the torrent status
        status_keys = self.status_keys
        if self.config['show_piecesbar']:
            status_keys.extend(['pieces', 'num_pieces'])

        component.get('SessionProxy').get_torrent_status(
            selected, status_keys
        ).addCallback(self._on_get_torrent_status)

    def _on_get_torrent_status(self, status):
        # Check to see if we got valid data from the core
        if not status:
            return

        # Update all the label widgets
        for widget in self.tab_widgets.values():
            txt = self.widget_status_as_fstr(widget, status)
            if decode_bytes(widget[0].get_text()) != txt:
                widget[0].set_text(txt)

        # Update progress bar separately as it's a special case (not a label).
        fraction = status['progress'] / 100

        if self.config['show_piecesbar']:
            if self.piecesbar.get_fraction() != fraction:
                self.piecesbar.set_fraction(fraction)
            if (
                status['state'] != 'Checking'
                and self.piecesbar.get_pieces() != status['pieces']
            ):
                # Skip pieces assignment if checking torrent.
                self.piecesbar.set_pieces(status['pieces'], status['num_pieces'])
            self.piecesbar.update()
        else:
            if self.progressbar.get_fraction() != fraction:
                self.progressbar.set_fraction(fraction)

    def on_show_piecesbar_config_changed(self, key, show):
        if show:
            self.show_piecesbar()
        else:
            self.hide_piecesbar()

    def show_piecesbar(self):
        if self.piecesbar is None:
            self.piecesbar = PiecesBar()
            self.main_builder.get_object('status_progress_vbox').pack_start(
                self.piecesbar, False, False, 0
            )
        self.tab_widgets['piecesbar'] = TabWidget(
            self.piecesbar, fpcnt, ('progress', 'state', 'message')
        )
        self.piecesbar.show()
        self.progressbar.hide()

    def hide_piecesbar(self):
        self.progressbar.show()
        if self.piecesbar:
            self.piecesbar.hide()
            self.tab_widgets.pop('piecesbar', None)
            self.piecesbar = None

    def clear(self):
        for widget in self.tab_widgets.values():
            widget[0].set_text('')

        if self.config['show_piecesbar']:
            self.piecesbar.clear()
        else:
            self.progressbar.set_fraction(0)



#
# Copyright (C) 2008 Andrew Resch <andrewresch@gmail.com>
#
# This file is part of Deluge and is licensed under GNU General Public License 3.0, or later, with
# the additional special exception to link portions of this program with the OpenSSL library.
# See LICENSE for more details.
#

from gi.repository.Gtk import IconSize

import deluge.common
import deluge.component as component
from deluge.configmanager import ConfigManager
from deluge.ui.client import client


class NewReleaseDialog:
    def __init__(self):
        pass

    def show(self, available_version):
        self.config = ConfigManager('gtk3ui.conf')
        main_builder = component.get('MainWindow').get_builder()
        self.dialog = main_builder.get_object('new_release_dialog')
        # Set the version labels
        if deluge.common.windows_check() or deluge.common.osx_check():
            main_builder.get_object('image_new_release').set_from_file(
                deluge.common.get_pixmap('deluge16.png')
            )
        else:
            main_builder.get_object('image_new_release').set_from_icon_name(
                'deluge', IconSize.LARGE_TOOLBAR
            )
        main_builder.get_object('label_available_version').set_text(available_version)
        main_builder.get_object('label_client_version').set_text(
            deluge.common.get_version()
        )
        self.chk_not_show_dialog = main_builder.get_object(
            'chk_do_not_show_new_release'
        )
        main_builder.get_object('button_goto_downloads').connect(
            'clicked', self._on_button_goto_downloads
        )
        main_builder.get_object('button_close_new_release').connect(
            'clicked', self._on_button_close_new_release
        )

        if client.connected():

            def on_info(version):
                main_builder.get_object('label_server_version').set_text(version)
                main_builder.get_object('label_server_version').show()
                main_builder.get_object('label_server_version_text').show()

            if not client.is_standalone():
                main_builder.get_object('label_client_version_text').set_label(
                    _('<i>Client Version</i>')
                )
                client.daemon.info().addCallback(on_info)

        self.dialog.show()

    def _on_button_goto_downloads(self, widget):
        deluge.common.open_url_in_browser('https://deluge-torrent.org')
        self.config['show_new_releases'] = not self.chk_not_show_dialog.get_active()
        self.dialog.destroy()

    def _on_button_close_new_release(self, widget):
        self.config['show_new_releases'] = not self.chk_not_show_dialog.get_active()
        self.dialog.destroy()



#
# Copyright (C) 2007, 2008 Andrew Resch <andrewresch@gmail.com>
#
# This file is part of Deluge and is licensed under GNU General Public License 3.0, or later, with
# the additional special exception to link portions of this program with the OpenSSL library.
# See LICENSE for more details.
#

"""The torrent view component that lists all torrents in the session."""

import logging
from locale import strcoll

from gi.repository.Gdk import ModifierType, keyval_name
from gi.repository.GLib import idle_add
from gi.repository.GObject import TYPE_UINT64
from gi.repository.Gtk import EntryIconPosition
from twisted.internet import reactor

import deluge.component as component
from deluge.common import decode_bytes
from deluge.ui.client import client

from . import torrentview_data_funcs as funcs
from .common import cmp
from .listview import ListView
from .removetorrentdialog import RemoveTorrentDialog

log = logging.getLogger(__name__)

try:
    CTRL_ALT_MASK = ModifierType.CONTROL_MASK | ModifierType.MOD1_MASK
except TypeError:
    # Sphinx AutoDoc has a mock issue with Gdk masks.
    pass


def str_nocase_sort(model, iter1, iter2, data):
    """Sort string column data using ISO 14651 in lowercase.

    Uses locale.strcoll which (allegedly) uses ISO 14651. Compares first
    value with second and returns -1, 0, 1 for where it should be placed.

    """
    v1 = model[iter1][data]
    v2 = model[iter2][data]
    # Catch any values of None from model.
    v1 = v1.lower() if v1 else ''
    v2 = v2.lower() if v2 else ''
    return strcoll(v1, v2)


def queue_peer_seed_sort_function(v1, v2):
    if v1 == v2:
        return 0
    if v2 < 0:
        return -1
    if v1 < 0:
        return 1
    if v1 > v2:
        return 1
    if v2 > v1:
        return -1


def queue_column_sort(model, iter1, iter2, data):
    v1 = model[iter1][data]
    v2 = model[iter2][data]
    return queue_peer_seed_sort_function(v1, v2)


def eta_column_sort(model, iter1, iter2, data):
    v1 = model[iter1][data]
    v2 = model[iter2][data]
    if v1 == v2:
        return 0
    if v1 == 0:
        return -1
    if v2 == 0:
        return 1
    if v1 > v2:
        return -1
    if v2 > v1:
        return 1


def seed_peer_column_sort(model, iter1, iter2, data):
    v1 = model[iter1][data]  # num seeds/peers
    v3 = model[iter2][data]  # num seeds/peers
    if v1 == v3:
        v2 = model[iter1][data + 1]  # total seeds/peers
        v4 = model[iter2][data + 1]  # total seeds/peers
        return queue_peer_seed_sort_function(v2, v4)
    return queue_peer_seed_sort_function(v1, v3)


def progress_sort(model, iter1, iter2, sort_column_id):
    progress1 = model[iter1][sort_column_id]
    progress2 = model[iter2][sort_column_id]
    # Progress value is equal, so sort on state
    if progress1 == progress2:
        state1 = model[iter1][sort_column_id + 1]
        state2 = model[iter2][sort_column_id + 1]
        return cmp(state1, state2)
    return cmp(progress1, progress2)


class SearchBox:
    def __init__(self, torrentview):
        self.torrentview = torrentview
        mainwindow = component.get('MainWindow')
        main_builder = mainwindow.get_builder()

        self.visible = False
        self.search_pending = self.prefiltered = None

        self.search_box = main_builder.get_object('search_box')
        self.search_torrents_entry = main_builder.get_object('search_torrents_entry')
        self.close_search_button = main_builder.get_object('close_search_button')
        self.match_search_button = main_builder.get_object('search_torrents_match')
        mainwindow.connect_signals(self)

    def show(self):
        self.visible = True
        self.search_box.show_all()
        self.search_torrents_entry.grab_focus()

    def hide(self):
        self.visible = False
        self.clear_search()
        self.search_box.hide()
        self.search_pending = self.prefiltered = None

    def clear_search(self):
        if self.search_pending and self.search_pending.active():
            self.search_pending.cancel()

        if self.prefiltered:
            filter_column = self.torrentview.columns['filter'].column_indices[0]
            torrent_id_column = self.torrentview.columns['torrent_id'].column_indices[0]
            for row in self.torrentview.liststore:
                torrent_id = row[torrent_id_column]

                if torrent_id in self.prefiltered:
                    # Reset to previous filter state
                    self.prefiltered.pop(self.prefiltered.index(torrent_id))
                    row[filter_column] = not row[filter_column]

        self.prefiltered = None

        self.search_torrents_entry.set_text('')
        if self.torrentview.filter and 'name' in self.torrentview.filter:
            self.torrentview.filter.pop('name', None)
            self.search_pending = reactor.callLater(0.5, self.torrentview.update)

    def set_search_filter(self):
        if self.search_pending and self.search_pending.active():
            self.search_pending.cancel()

        if self.torrentview.filter and 'name' in self.torrentview.filter:
            self.torrentview.filter.pop('name', None)

        elif self.torrentview.filter is None:
            self.torrentview.filter = {}

        search_string = self.search_torrents_entry.get_text()
        if not search_string:
            self.clear_search()
        else:
            if self.match_search_button.get_active():
                search_string += '::match'
            self.torrentview.filter['name'] = search_string
        self.prefilter_torrentview()

    def prefilter_torrentview(self):
        filter_column = self.torrentview.columns['filter'].column_indices[0]
        torrent_id_column = self.torrentview.columns['torrent_id'].column_indices[0]
        torrent_name_column = self.torrentview.columns[_('Name')].column_indices[1]

        match_case = self.match_search_button.get_active()
        if match_case:
            search_string = self.search_torrents_entry.get_text()
        else:
            search_string = self.search_torrents_entry.get_text().lower()

        if self.prefiltered is None:
            self.prefiltered = []

        for row in self.torrentview.liststore:
            torrent_id = row[torrent_id_column]

            if torrent_id in self.prefiltered:
                # Reset to previous filter state
                self.prefiltered.pop(self.prefiltered.index(torrent_id))
                row[filter_column] = not row[filter_column]

            if not row[filter_column]:
                # Row is not visible(filtered out, but not by our filter), skip it
                continue

            if match_case:
                torrent_name = row[torrent_name_column]
            else:
                torrent_name = row[torrent_name_column].lower()

            if search_string in torrent_name and not row[filter_column]:
                row[filter_column] = True
                self.prefiltered.append(torrent_id)
            elif search_string not in torrent_name and row[filter_column]:
                row[filter_column] = False
                self.prefiltered.append(torrent_id)

    def on_close_search_button_clicked(self, widget):
        self.hide()

    def on_search_filter_toggle(self, widget):
        if self.visible:
            self.hide()
        else:
            self.show()

    def on_search_torrents_match_toggled(self, widget):
        if self.search_torrents_entry.get_text():
            self.set_search_filter()
            self.search_pending = reactor.callLater(0.7, self.torrentview.update)

    def on_search_torrents_entry_icon_press(self, entry, icon, event):
        if icon != EntryIconPosition.SECONDARY:
            return
        self.clear_search()

    def on_search_torrents_entry_changed(self, widget):
        self.set_search_filter()
        self.search_pending = reactor.callLater(0.7, self.torrentview.update)


class TorrentView(ListView, component.Component):
    """TorrentView handles the listing of torrents."""

    def __init__(self):
        component.Component.__init__(
            self, 'TorrentView', interval=2, depend=['SessionProxy']
        )
        main_builder = component.get('MainWindow').get_builder()
        # Call the ListView constructor
        ListView.__init__(
            self, main_builder.get_object('torrent_view'), 'torrentview.state'
        )
        log.debug('TorrentView Init..')

        # If we have gotten the state yet
        self.got_state = False

        # This is where status updates are put
        self.status = {}

        # We keep a copy of the previous status to compare for changes
        self.prev_status = {}

        # Register the columns menu with the listview so it gets updated accordingly.
        self.register_checklist_menu(main_builder.get_object('menu_columns'))

        # Add the columns to the listview
        self.add_text_column('torrent_id', hidden=True, unique=True)
        self.add_bool_column('dirty', hidden=True)
        self.add_func_column(
            '#',
            funcs.cell_data_queue,
            [int],
            status_field=['queue'],
            sort_func=queue_column_sort,
        )
        self.add_texticon_column(
            _('Name'),
            status_field=['state', 'name'],
            function=funcs.cell_data_statusicon,
            sort_func=str_nocase_sort,
            default_sort=True,
        )
        self.add_func_column(
            _('Size'),
            funcs.cell_data_size,
            [TYPE_UINT64],
            status_field=['total_wanted'],
        )
        self.add_func_column(
            _('Downloaded'),
            funcs.cell_data_size,
            [TYPE_UINT64],
            status_field=['all_time_download'],
            default=False,
        )
        self.add_func_column(
            _('Uploaded'),
            funcs.cell_data_size,
            [TYPE_UINT64],
            status_field=['total_uploaded'],
            default=False,
        )
        self.add_func_column(
            _('Remaining'),
            funcs.cell_data_size,
            [TYPE_UINT64],
            status_field=['total_remaining'],
            default=False,
        )
        self.add_progress_column(
            _('Progress'),
            status_field=['progress', 'state'],
            col_types=[float, str],
            function=funcs.cell_data_progress,
            sort_func=progress_sort,
        )
        self.add_func_column(
            _('Seeds'),
            funcs.cell_data_peer,
            [int, int],
            status_field=['num_seeds', 'total_seeds'],
            sort_func=seed_peer_column_sort,
            default=False,
        )
        self.add_func_column(
            _('Peers'),
            funcs.cell_data_peer,
            [int, int],
            status_field=['num_peers', 'total_peers'],
            sort_func=seed_peer_column_sort,
            default=False,
        )
        self.add_func_column(
            _('Seeds:Peers'),
            funcs.cell_data_ratio_seeds_peers,
            [float],
            status_field=['seeds_peers_ratio'],
            default=False,
        )
        self.add_func_column(
            _('Down Speed'),
            funcs.cell_data_speed_down,
            [int],
            status_field=['download_payload_rate'],
        )
        self.add_func_column(
            _('Up Speed'),
            funcs.cell_data_speed_up,
            [int],
            status_field=['upload_payload_rate'],
        )
        self.add_func_column(
            _('Down Limit'),
            funcs.cell_data_speed_limit_down,
            [float],
            status_field=['max_download_speed'],
            default=False,
        )
        self.add_func_column(
            _('Up Limit'),
            funcs.cell_data_speed_limit_up,
            [float],
            status_field=['max_upload_speed'],
            default=False,
        )
        self.add_func_column(
            _('ETA'),
            funcs.cell_data_time,
            [int],
            status_field=['eta'],
            sort_func=eta_column_sort,
        )
        self.add_func_column(
            _('Ratio'),
            funcs.cell_data_ratio_ratio,
            [float],
            status_field=['ratio'],
            default=False,
        )
        self.add_func_column(
            _('Avail'),
            funcs.cell_data_ratio_avail,
            [float],
            status_field=['distributed_copies'],
            default=False,
        )
        self.add_func_column(
            _('Added'),
            funcs.cell_data_date_added,
            [int],
            status_field=['time_added'],
            default=False,
        )
        self.add_func_column(
            _('Completed'),
            funcs.cell_data_date_completed,
            [int],
            status_field=['completed_time'],
            default=False,
        )
        self.add_func_column(
            _('Complete Seen'),
            funcs.cell_data_date_or_never,
            [int],
            status_field=['last_seen_complete'],
            default=False,
        )
        self.add_func_column(
            _('Last Transfer'),
            funcs.cell_data_time,
            [int],
            status_field=['time_since_transfer'],
            default=False,
        )
        self.add_texticon_column(
            _('Tracker'),
            function=funcs.cell_data_trackericon,
            status_field=['tracker_host', 'tracker_host'],
            default=False,
        )
        self.add_text_column(
            _('Download Folder'), status_field=['download_location'], default=False
        )
        self.add_text_column(_('Owner'), status_field=['owner'], default=False)
        self.add_bool_column(
            _('Shared'),
            status_field=['shared'],
            default=False,
            tooltip=_('Torrent is shared between other Deluge users or not.'),
        )
        self.restore_columns_order_from_state()

        # Set filter to None for now
        self.filter = None

        # Connect Signals #
        # Connect to the 'button-press-event' to know when to bring up the
        # torrent menu popup.
        self.treeview.connect('button-press-event', self.on_button_press_event)
        # Connect to the 'key-press-event' to know when the bring up the
        # torrent menu popup via keypress.
        self.treeview.connect('key-release-event', self.on_key_press_event)
        # Connect to the 'changed' event of TreeViewSelection to get selection
        # changes.
        self.treeview.get_selection().connect('changed', self.on_selection_changed)

        self.treeview.connect('drag-drop', self.on_drag_drop)
        self.treeview.connect('drag_data_received', self.on_drag_data_received)
        self.treeview.connect('key-press-event', self.on_key_press_event)
        self.treeview.connect('columns-changed', self.on_columns_changed_event)

        self.search_box = SearchBox(self)
        self.permanent_status_keys = ['owner']
        self.columns_to_update = []

    def start(self):
        """Start the torrentview"""
        # We need to get the core session state to know which torrents are in
        # the session so we can add them to our list.
        # Only get the status fields required for the visible columns
        status_fields = []
        for listview_column in self.columns.values():
            if listview_column.column.get_visible():
                if not listview_column.status_field:
                    continue
                status_fields.extend(listview_column.status_field)
        component.get('SessionProxy').get_torrents_status(
            {}, status_fields
        ).addCallback(self._on_session_state)

        client.register_event_handler(
            'TorrentStateChangedEvent', self.on_torrentstatechanged_event
        )
        client.register_event_handler('TorrentAddedEvent', self.on_torrentadded_event)
        client.register_event_handler(
            'TorrentRemovedEvent', self.on_torrentremoved_event
        )
        client.register_event_handler('SessionPausedEvent', self.on_sessionpaused_event)
        client.register_event_handler(
            'SessionResumedEvent', self.on_sessionresumed_event
        )
        client.register_event_handler(
            'TorrentQueueChangedEvent', self.on_torrentqueuechanged_event
        )

    def _on_session_state(self, state):
        self.add_rows(state)
        self.got_state = True
        # Update the view right away with our status
        self.status = state
        self.set_columns_to_update()
        self.update_view(load_new_list=True)
        self.select_first_row()

    def stop(self):
        """Stops the torrentview"""
        client.deregister_event_handler(
            'TorrentStateChangedEvent', self.on_torrentstatechanged_event
        )
        client.deregister_event_handler('TorrentAddedEvent', self.on_torrentadded_event)
        client.deregister_event_handler(
            'TorrentRemovedEvent', self.on_torrentremoved_event
        )
        client.deregister_event_handler(
            'SessionPausedEvent', self.on_sessionpaused_event
        )
        client.deregister_event_handler(
            'SessionResumedEvent', self.on_sessionresumed_event
        )
        client.deregister_event_handler(
            'TorrentQueueChangedEvent', self.on_torrentqueuechanged_event
        )

        if self.treeview.get_selection():
            self.treeview.get_selection().unselect_all()

        # Save column state before clearing liststore
        # so column sort details are correctly saved.
        self.save_state()
        self.liststore.clear()
        self.prev_status = {}
        self.filter = None
        self.search_box.hide()

    def shutdown(self):
        """Called when GtkUi is exiting"""
        pass

    def save_state(self):
        """
        Saves the state of the torrent view.
        """
        if component.get('MainWindow').visible():
            ListView.save_state(self, 'torrentview.state')

    def remove_column(self, header):
        """Removes the column with the name 'header' from the torrentview"""
        self.save_state()
        ListView.remove_column(self, header)

    def set_filter(self, filter_dict):
        """
        Sets filters for the torrentview..

        see: core.get_torrents_status
        """
        search_filter = self.filter and self.filter.get('name', None) or None
        self.filter = dict(filter_dict)  # Copied version of filter_dict.
        if search_filter and 'name' not in filter_dict:
            self.filter['name'] = search_filter
        self.update(select_row=True)

    def set_columns_to_update(self, columns=None):
        status_keys = []
        self.columns_to_update = []

        if columns is None:
            # We need to iterate through all columns
            columns = list(self.columns)

        # Iterate through supplied list of columns to update
        for column in columns:
            # Make sure column is visible and has 'status_field' set.
            # If not, we can ignore it.
            if (
                self.columns[column].column.get_visible() is True
                and self.columns[column].hidden is False
                and self.columns[column].status_field is not None
            ):
                for field in self.columns[column].status_field:
                    status_keys.append(field)
                    self.columns_to_update.append(column)

        # Remove duplicates
        self.columns_to_update = list(set(self.columns_to_update))
        status_keys = list(set(status_keys + self.permanent_status_keys))
        return status_keys

    def send_status_request(self, columns=None, select_row=False):
        # Store the 'status_fields' we need to send to core
        status_keys = self.set_columns_to_update(columns)

        # If there is nothing in status_keys then we must not continue
        if status_keys == []:
            return

        # Remove duplicates from status_key list
        status_keys = list(set(status_keys))

        # Request the statuses for all these torrent_ids, this is async so we
        # will deal with the return in a signal callback.
        d = (
            component.get('SessionProxy')
            .get_torrents_status(self.filter, status_keys)
            .addCallback(self._on_get_torrents_status)
        )
        if select_row:
            d.addCallback(self.select_first_row)

    def select_first_row(self, ignored=None):
        """
        Set the first row in the list selected if a selection does
        not already exist
        """
        rows = self.treeview.get_selection().get_selected_rows()[1]
        # Only select row if noe rows are selected
        if not rows:
            self.treeview.get_selection().select_path((0,))

    def update(self, select_row=False):
        """
        Sends a status request to core and updates the torrent list with the result.

        :param select_row: if the first row in the list should be selected if
                           no rows are already selected.
        :type select_row: boolean

        """
        if self.got_state:
            if (
                self.search_box.search_pending is not None
                and self.search_box.search_pending.active()
            ):
                # An update request is scheduled, let's wait for that one
                return
            # Send a status request
            idle_add(self.send_status_request, None, select_row)

    def update_view(self, load_new_list=False):
        """Update the torrent view model with data we've received."""
        filter_column = self.columns['filter'].column_indices[0]
        status = self.status

        if not load_new_list:
            # Freeze notications while updating
            self.treeview.freeze_child_notify()

        # Get the columns to update from one of the torrents
        if status:
            torrent_id = list(status)[0]
            fields_to_update = []
            for column in self.columns_to_update:
                column_index = self.get_column_index(column)
                for i, status_field in enumerate(self.columns[column].status_field):
                    # Only use columns that the torrent has in the state
                    if status_field in status[torrent_id]:
                        fields_to_update.append((column_index[i], status_field))

        for row in self.liststore:
            torrent_id = row[self.columns['torrent_id'].column_indices[0]]
            # We expect the torrent_id to be in status and prev_status,
            # as it will be as long as the list isn't changed by the user

            torrent_id_in_status = False
            try:
                torrent_status = status[torrent_id]
                torrent_id_in_status = True
                if torrent_status == self.prev_status[torrent_id]:
                    # The status dict is the same, so do nothing to update for this torrent
                    continue
            except KeyError:
                pass

            if not torrent_id_in_status:
                if row[filter_column] is True:
                    row[filter_column] = False
            else:
                if row[filter_column] is False:
                    row[filter_column] = True

                # Find the fields to update
                to_update = []
                for i, status_field in fields_to_update:
                    row_value = status[torrent_id][status_field]
                    if decode_bytes(row[i]) != row_value:
                        to_update.append(i)
                        to_update.append(row_value)
                # Update fields in the liststore
                if to_update:
                    self.liststore.set(row.iter, *to_update)

        if load_new_list:
            # Create the model filter. This sets the model for the treeview and enables sorting.
            self.create_model_filter()
        else:
            self.treeview.thaw_child_notify()

        component.get('MenuBar').update_menu()
        self.prev_status = status

    def _on_get_torrents_status(self, status, select_row=False):
        """Callback function for get_torrents_status().  'status' should be a
        dictionary of {torrent_id: {key, value}}."""
        self.status = status
        if self.search_box.prefiltered is not None:
            self.search_box.prefiltered = None

        if self.status == self.prev_status and self.prev_status:
            # We do not bother updating since the status hasn't changed
            self.prev_status = self.status
            return
        self.update_view()

    def add_rows(self, torrent_ids):
        """Accepts a list of torrent_ids to add to self.liststore"""
        torrent_id_column = self.columns['torrent_id'].column_indices[0]
        dirty_column = self.columns['dirty'].column_indices[0]
        filter_column = self.columns['filter'].column_indices[0]
        for torrent_id in torrent_ids:
            # Insert a new row to the liststore
            row = self.liststore.append()
            self.liststore.set(
                row,
                torrent_id_column,
                torrent_id,
                dirty_column,
                True,
                filter_column,
                True,
            )

    def remove_row(self, torrent_id):
        """Removes a row with torrent_id"""
        for row in self.liststore:
            if row[self.columns['torrent_id'].column_indices[0]] == torrent_id:
                self.liststore.remove(row.iter)
                # Force an update of the torrentview
                self.update(select_row=True)
                break

    def mark_dirty(self, torrent_id=None):
        for row in self.liststore:
            if (
                not torrent_id
                or row[self.columns['torrent_id'].column_indices[0]] == torrent_id
            ):
                # log.debug('marking %s dirty', torrent_id)
                row[self.columns['dirty'].column_indices[0]] = True
                if torrent_id:
                    break

    def get_selected_torrent(self):
        """Returns a torrent_id or None.  If multiple torrents are selected,
        it will return the torrent_id of the first one."""
        selected = self.get_selected_torrents()
        if selected:
            return selected[0]
        else:
            return selected

    def get_selected_torrents(self):
        """Returns a list of selected torrents or None"""
        torrent_ids = []
        try:
            paths = self.treeview.get_selection().get_selected_rows()[1]
        except AttributeError:
            # paths is likely None .. so lets return []
            return []
        try:
            for path in paths:
                try:
                    row = self.treeview.get_model().get_iter(path)
                except Exception as ex:
                    log.debug('Unable to get iter from path: %s', ex)
                    continue

                child_row = self.treeview.get_model().convert_iter_to_child_iter(row)
                child_row = (
                    self.treeview.get_model()
                    .get_model()
                    .convert_iter_to_child_iter(child_row)
                )
                if self.liststore.iter_is_valid(child_row):
                    try:
                        value = self.liststore.get_value(
                            child_row, self.columns['torrent_id'].column_indices[0]
                        )
                    except Exception as ex:
                        log.debug('Unable to get value from row: %s', ex)
                    else:
                        torrent_ids.append(value)
            if len(torrent_ids) == 0:
                return []

            return torrent_ids
        except (ValueError, TypeError):
            return []

    def get_torrent_status(self, torrent_id):
        """Returns data stored in self.status, it may not be complete"""
        try:
            return self.status[torrent_id]
        except KeyError:
            return {}

    def get_visible_torrents(self):
        return list(self.status)

    # Callbacks #
    def on_button_press_event(self, widget, event):
        """This is a callback for showing the right-click context menu."""
        log.debug('on_button_press_event')
        # We only care about right-clicks
        if event.button == 3 and event.window == self.treeview.get_bin_window():
            x, y = event.get_coords()
            path = self.treeview.get_path_at_pos(int(x), int(y))
            if not path:
                return
            row = self.model_filter.get_iter(path[0])

            if self.get_selected_torrents():
                if (
                    self.model_filter.get_value(
                        row, self.columns['torrent_id'].column_indices[0]
                    )
                    not in self.get_selected_torrents()
                ):
                    self.treeview.get_selection().unselect_all()
                    self.treeview.get_selection().select_iter(row)
            else:
                self.treeview.get_selection().select_iter(row)
            torrentmenu = component.get('MenuBar').torrentmenu
            torrentmenu.popup(None, None, None, None, event.button, event.time)
            return True

    def on_selection_changed(self, treeselection):
        """This callback is know when the selection has changed."""
        log.debug('on_selection_changed')
        component.get('TorrentDetails').update()
        component.get('MenuBar').update_menu()

    def on_drag_drop(self, widget, drag_context, x, y, timestamp):
        widget.stop_emission('drag-drop')

    def on_drag_data_received(
        self, widget, drag_context, x, y, selection_data, info, timestamp
    ):
        widget.stop_emission('drag_data_received')

    def on_columns_changed_event(self, treeview):
        log.debug('Treeview Columns Changed')
        self.save_state()

    def on_torrentadded_event(self, torrent_id, from_state):
        self.add_rows([torrent_id])
        self.update(select_row=True)

    def on_torrentremoved_event(self, torrent_id):
        self.remove_row(torrent_id)

    def on_torrentstatechanged_event(self, torrent_id, state):
        # Update the torrents state
        for row in self.liststore:
            if torrent_id != row[self.columns['torrent_id'].column_indices[0]]:
                continue

            for name in self.columns_to_update:
                if not self.columns[name].status_field:
                    continue
                for idx, status_field in enumerate(self.columns[name].status_field):
                    # Update all columns that use the state field to current state
                    if status_field != 'state':
                        continue
                    row[self.get_column_index(name)[idx]] = state

            if self.filter.get('state', None) is not None:
                # We have a filter set, let's see if theres anything to hide
                # and remove from status
                if (
                    torrent_id in self.status
                    and self.status[torrent_id]['state'] != state
                ):
                    row[self.columns['filter'].column_indices[0]] = False
                    del self.status[torrent_id]

        self.mark_dirty(torrent_id)

    def on_sessionpaused_event(self):
        self.mark_dirty()
        self.update()

    def on_sessionresumed_event(self):
        self.mark_dirty()
        self.update(select_row=True)

    def on_torrentqueuechanged_event(self):
        self.mark_dirty()
        self.update()

    # Handle keyboard shortcuts
    def on_key_press_event(self, widget, event):
        keyname = keyval_name(event.keyval)
        if keyname is not None:
            func = getattr(self, 'keypress_' + keyname.lower(), None)
            if func:
                return func(event)

    def keypress_up(self, event):
        """Handle any Up arrow keypresses"""
        log.debug('keypress_up')
        torrents = self.get_selected_torrents()
        if not torrents:
            return

        # Move queue position up with Ctrl+Alt or Ctrl+Alt+Shift
        if event.get_state() & CTRL_ALT_MASK:
            if event.get_state() & ModifierType.SHIFT_MASK:
                client.core.queue_top(torrents)
            else:
                client.core.queue_up(torrents)

    def keypress_down(self, event):
        """Handle any Down arrow keypresses"""
        log.debug('keypress_down')
        torrents = self.get_selected_torrents()
        if not torrents:
            return

        # Move queue position down with Ctrl+Alt or Ctrl+Alt+Shift
        if event.get_state() & CTRL_ALT_MASK:
            if event.get_state() & ModifierType.SHIFT_MASK:
                client.core.queue_bottom(torrents)
            else:
                client.core.queue_down(torrents)

    def keypress_delete(self, event):
        log.debug('keypress_delete')
        torrents = self.get_selected_torrents()
        if torrents:
            if event.get_state() & ModifierType.SHIFT_MASK:
                RemoveTorrentDialog(torrents, delete_files=True).run()
            else:
                RemoveTorrentDialog(torrents).run()

    def keypress_menu(self, event):
        log.debug('keypress_menu')
        if not self.get_selected_torrent():
            return

        torrentmenu = component.get('MenuBar').torrentmenu
        torrentmenu.popup(None, None, None, None, 3, event.time)
        return True
